#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from io import StringIO
from os.path import basename, normpath
from typing import Any, Collection, Dict, List, Optional, Tuple

import pandas as pd

from mitosheet.code_chunks.step_performers.import_steps.simple_import_code_chunk import \
    generate_read_csv_code
from mitosheet.preprocessing.preprocess_step_performer import \
    PreprocessStepPerformer
from mitosheet.step_performers.import_steps.simple_import import \
    read_csv_get_delimiter_and_encoding
from mitosheet.telemetry.telemetry_utils import log
from mitosheet.transpiler.transpile_utils import get_column_header_as_transpiled_code, get_column_header_list_as_transpiled_code, get_str_param_name
from mitosheet.types import StepsManagerType
from mitosheet.utils import get_valid_dataframe_name

def convert_arg_of_string_type_to_dataframe(arg: str, arg_index: int) -> Optional[Tuple[Any, pd.DataFrame, str, str, Dict[str, Any]]]:
    """
    Converts a string to a dataframe if possible. If not possible, then returns None.
    """
    # First, try to read in the file path
    try:
        # We use the simple import 
        df, delimeter, encoding = read_csv_get_delimiter_and_encoding(arg)
        return arg, df, basename(normpath(arg)), 'file_path', {
            'delimeter': delimeter,
            'encoding': encoding
        }
    except:
        pass

    # If we can't read in the file path, then it's possible that this is a CSV string of a dataframe
    # in which case, we can read it in with StringIO
    if not arg.startswith('{'): # Skip this as this means it's to_json converted
        try:
            df = pd.read_csv(StringIO(arg))
            return arg, df, 'df' + str(arg_index + 1), 'csv_string', {}
        except:
            pass
            
    # If we still can't read that, then try this as a JSON string
    try:
        df = pd.read_json(arg)
        return arg, df, 'df' + str(arg_index + 1), 'json_string', {}
    except:
        pass

    return None

def convert_arg_of_unknown_type_to_dataframe(arg: Any, arg_index: int) -> Optional[List[Tuple[Any, pd.DataFrame, str, str, Dict[str, Any]]]]:
    """
    Accepts the following arguments:
    1. dataframe: A dataframe
    2. file_path: A string that is a path to a CSV file
    3. csv_string: A string that is a CSV of a dataframe
    4. json_string: A string that is the JSON of a dataframe
    5. to_dict_records: A list that is the .to_dict('records') of a dataframe
    6. list_of_csv_strings: A list of strings that are CSVs of dataframes

    TODO: 
    - series: A series
    - Add support for Excel files
    - Add support for JSON files
    - Add support for URLs that have multiple tables on them
    - Add support for URLs that have a single table on them
    - Add support for Conn objects?
    - Add support for other types of dataframes?

    Returns a list of tuples of the dataframe, dataframe name, type being converted from, and extra data
    """
    if isinstance(arg, pd.DataFrame):
        return [(arg, arg, 'df' + str(arg_index + 1), 'dataframe', {})]
    
    if isinstance(arg, str):
        result = convert_arg_of_string_type_to_dataframe(arg, arg_index)
        if result is not None:
            return [result]
        return None

    if isinstance(arg, list):
        # If this is a list, then it could be the .to_dict('records') of a dataframe
        if len(arg) == 0:
            return None
        
        if isinstance(arg[0], dict):
            try:
                # If this is a list of dicts, then it could be the .to_dict('records') of a dataframe
                return [(arg, pd.DataFrame(arg), 'df' + str(arg_index + 1), 'to_dict_records', {})]
            except:
                pass
        
        if isinstance(arg[0], str):
            # If this is a list of strings, then it could be a list of CSV strings
            results = []
            for arg_index, arg in enumerate(arg):
                result = convert_arg_of_string_type_to_dataframe(arg, arg_index)
                if result is None:
                    return None
                results.append(result)
            return results

    return None


class ConvertToDataframePreprocessStepPerformer(PreprocessStepPerformer):
    """
    This preprocessing step is responsible for converting
    all of the args to dataframes.

    If it fails to convert any of the arguments to a dataframe,
    then it will throw an error.
    """

    @classmethod
    def preprocess_step_version(cls) -> int:
        return 1

    @classmethod
    def preprocess_step_type(cls) -> str:
        return 'convert_to_dataframe'

    @classmethod
    def execute(cls, args: Collection[Any]) -> Tuple[List[Any], Optional[List[str]], Optional[Dict[str, Any]]]:
        df_args: List[pd.DataFrame] = []
        df_names: List[str] = []

        # Keep all the information we need to convert each arg, 
        # A list of (original_arg, df_name, type, conversion_data)
        conversion_params: List[Tuple[Any, str, str, Dict[str, Any]]] = []

        for arg_index, arg in enumerate(args):
            results = convert_arg_of_unknown_type_to_dataframe(arg, arg_index)
            if results is None:
                error_message = f'Invalid argument passed to sheet: {arg}. Please pass all dataframes, paths to CSV files, or types that can be converted to Pandas dataframes.'
                log('mitosheet_sheet_call_failed', {'error': error_message}, failed=True)
                raise ValueError(error_message)

            for result in results:
                original_arg, df, df_name, type_converted_from, extra_data = result
                # If the dataframe name is already taken, then we need to change it
                df_name = get_valid_dataframe_name(df_names, df_name)
                df_args.append(df)
                df_names.append(df_name)
                conversion_params.append((original_arg, df_name, type_converted_from, extra_data))                
                
        return df_args, df_names, {
            'df_names': df_names,
            'conversion_params': conversion_params
        }

    @classmethod
    def transpile(cls, steps_manager: StepsManagerType, execution_data: Optional[Dict[str, Any]]) -> Tuple[List[str], List[str]]:
        """
        Transpiles the reading in of passed file paths to dataframe names, 
        with a simple pd.read_csv call.
        """
        code = []
        imports = []
        
        conversion_params = execution_data['conversion_params'] if execution_data is not None else []

        for arg_index, (arg, df_name, type_converted_from, extra_data) in enumerate(conversion_params):
            if type_converted_from == 'dataframe':
                # If this is a dataframe, then we don't need to do anything
                continue
            elif type_converted_from == 'series':
                imports += ['import pandas as pd']
                # If this is a series, then we can convert it to a dataframe
                code.append(
                    f'{df_name} = pd.DataFrame({arg})'
                )
            elif type_converted_from == 'file_path':
                imports += ['import pandas as pd']

                # Make sure to compile the path as a variable if the user is creating a function
                # TODO: fix this arg_index in the case that someone passes a list. Luckily, this 
                # rarely happens, but you can see the issue here: https://github.com/mito-ds/mito/issues/985
                file_name: str = arg if not steps_manager.code_options['as_function'] else get_str_param_name(steps_manager, arg_index) # type: ignore
                delimeter = extra_data['delimeter']
                encoding = extra_data['encoding']
                read_csv_code = generate_read_csv_code(
                    file_name, df_name, delimeter, encoding, None, None, None, file_name_is_variable=steps_manager.code_options['as_function']
                )

                code.append(
                    read_csv_code
                )
            elif type_converted_from == 'csv_string':
                imports += ['from io import StringIO']
                # If this is a CSV string, then we can read it in with StringIO
                code.append(
                    f'{df_name} = pd.read_csv(StringIO({get_column_header_as_transpiled_code(arg)}))'
                )
            elif type_converted_from == 'json_string':
                imports += ['import pandas as pd']
                code.append(
                    f'{df_name} = pd.read_json({get_column_header_as_transpiled_code(arg)})'
                )
            elif type_converted_from == 'to_dict_records':
                imports += ['import pandas as pd']
                code.append(
                    f'{df_name} = pd.DataFrame({get_column_header_list_as_transpiled_code(arg)})'
                )
            else:
                raise ValueError(f'Unknown type converted from: {type_converted_from}')

        if len(code) > 0:
            code.insert(0, '# Read in filepaths as dataframes')
                
        return code, imports


import json
import os
from typing import Dict

import requests


def create_github_pr(
        github_repo: str, 
        new_branch_name: str, 
        commit_message: str,
        pr_title: str, 
        pr_description: str, 
        new_files: Dict[str, str]
    ) -> str:
    """
    Create a new GitHub PR with the given parameters.
    """

    token = os.environ.get('GITHUB_TOKEN')
    if not token:
        raise Exception("GitHub token not found in environment variables. See docs here: TODO: insert link")

    headers = {
        'Authorization': f'token {token}',
        'Accept': 'application/vnd.github.v3+json',
    }

    api_url = f'https://api.github.com/repos/{github_repo}'

    # Check if the branch already exists
    existing_branch = requests.get(f'{api_url}/git/refs/heads/{new_branch_name}', headers=headers)
    if existing_branch.status_code == 200:
        raise Exception(f"An automation with this name already exists. Please pick a different name.")

    # Get the default branch of the repo
    repo_info = requests.get(api_url, headers=headers).json()
    default_branch = repo_info['default_branch']
    sha_latest_commit = requests.get(f'{api_url}/git/refs/heads/{default_branch}', headers=headers).json()['object']['sha']    

    # Create a new branch
    new_branch_data = {
        'ref': f'refs/heads/{new_branch_name}',
        'sha': sha_latest_commit
    }
    requests.post(f'{api_url}/git/refs', headers=headers, data=json.dumps(new_branch_data))

    # Get the current tree of the latest commit to preserve existing files
    latest_commit = requests.get(f'{api_url}/git/commits/{sha_latest_commit}', headers=headers).json()
    base_tree_sha = latest_commit['tree']['sha']

    tree_elements = []
    for file_path, file_content in new_files.items():
        tree_elements.append({
            'path': file_path,
            'mode': '100644',  # Blob (file) mode
            'type': 'blob',
            'content': file_content
        })

    tree_data = {
        'tree': tree_elements,
        'base_tree': base_tree_sha  # This is important to preserve existing files
    }
    tree_response = requests.post(f'{api_url}/git/trees', headers=headers, data=json.dumps(tree_data))

    if tree_response.status_code not in [200, 201]:
        raise Exception(f"Failed to create tree: {tree_response.json()}")

    tree_sha = tree_response.json()['sha']

    # Create a commit with the new tree
    commit_data = {
        'message': commit_message,
        'parents': [sha_latest_commit],
        'tree': tree_sha
    }
    commit_response = requests.post(f'{api_url}/git/commits', headers=headers, data=json.dumps(commit_data))

    if commit_response.status_code not in [200, 201]:
        raise Exception(f"Failed to create commit: {commit_response.json()}")

    commit_sha = commit_response.json()['sha']

    # Update the branch reference to point to the new commit
    update_branch_data = {
        'sha': commit_sha,
        'force': False
    }
    update_branch = requests.patch(f'{api_url}/git/refs/heads/{new_branch_name}', headers=headers, data=json.dumps(update_branch_data))
    if update_branch.status_code not in [200, 201]:
        raise Exception(f"Failed to update branch: {update_branch.json()}")

    # Create a pull request
    pr_data = {
        'title': pr_title,
        'head': new_branch_name,
        'base': default_branch,
        'body': pr_description
    }
    pr_response = requests.post(f'{api_url}/pulls', headers=headers, data=json.dumps(pr_data))

    if pr_response.status_code == 201:
        return pr_response.json()['html_url']
    else:
        raise Exception(f"Failed to create PR: {pr_response.content.decode('utf-8')}")



import importlib
import inspect
import os
from typing import Any, Callable, Dict, List, Optional, Tuple

import pandas as pd

from mitosheet.errors import MitoError
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import \
    get_column_header_as_transpiled_code
from mitosheet.types import UserDefinedFunctionParamType
from mitosheet.extensions.v1 import ColumnHeader


def check_valid_sheet_functions(
        sheet_functions: Optional[List[Callable]]=None,
    ) -> None:
    if sheet_functions is None or len(sheet_functions) == 0:
        return

    from mitosheet.user.utils import is_enterprise, is_running_test
    if not is_enterprise() and not is_running_test():
        raise ValueError("sheet_functions are only supported in the enterprise version of Mito. See Mito plans https://www.trymito.io/plans")

    if not isinstance(sheet_functions, list):
        raise ValueError(f"sheet_functions must be a list, but got {type(sheet_functions)}")
    
    for sheet_function in sheet_functions:
        if not callable(sheet_function):
            raise ValueError(f"sheet_functions must be a list of functions, but got {sheet_function} which is not callable.")
        
        # Check if has a __name__ attribute
        if not hasattr(sheet_function, '__name__'):
            raise ValueError(f"sheet_functions must be a list of functions, but got {sheet_function} which does not have a __name__ attribute. Please use a named function instead.")
        
        if sheet_function.__name__ == '<lambda>':
            raise ValueError(f"sheet_functions must be a list of functions, but got {sheet_function} which is a lambda function. Please use a named function instead.")
        
        # Check the name is all caps
        if not sheet_function.__name__.isupper():
            raise ValueError(f"sheet_functions must be a list of functions, but got {sheet_function} which has a name that is not all caps. Please use a named function instead.")
    
def get_functions_from_path(path: str) -> List[Callable]:
    # Get the filename from the path
    filename = os.path.basename(path)

    # Create a module spec from the file path
    module_spec = importlib.util.spec_from_file_location(filename, path)

    if module_spec is None:
        # Does path end in .py
        if not path.endswith('.py'):
            raise ValueError(f"Could not find a module spec for {path}. The path must end in .py for custom sheet functions to be loaded. Please have an admin update this file path and try again.")
        raise ValueError(f"Could not find a module spec for {path}. Please have an admin update this file path and try again.")

    # Create the module by loading the spec
    custom_functions_module = importlib.util.module_from_spec(module_spec)

    try:
        # Execute the module (this runs the code in the file)
        module_spec.loader.exec_module(custom_functions_module) # type: ignore
    except Exception:
        raise ImportError(f"The file path {path} does not exist, and so this file cannot be read in for the custom sheet functions. Please have an admin update this file path and try again.")
    
    # Get a list of functions defined in custom functions file
    functions = [getattr(custom_functions_module, attr) for attr in dir(custom_functions_module) if callable(getattr(custom_functions_module, attr))]

    # Filter out private functions
    functions = [func for func in functions if not func.__name__.startswith('_')]

    # Filter out functions imported from a different modules
    functions = [func for func in functions if func.__module__ == custom_functions_module.__name__]

    # Now, function_list contains the callable objects (functions) defined in custom functions file
    return functions


def get_non_validated_custom_sheet_functions(path: str) -> List[Callable]:
    functions = get_functions_from_path(path)

    # Filter out any functions that are not all uppercase
    return [func for func in functions if func.__name__.isupper()]


def validate_and_wrap_sheet_functions(user_defined_sheet_functions: Optional[List[Callable]]) -> List[Callable]:
    check_valid_sheet_functions(user_defined_sheet_functions)
    from mitosheet.public.v3.errors import handle_sheet_function_errors
    user_defined_functions = [handle_sheet_function_errors(user_defined_sheet_function) for user_defined_sheet_function in (user_defined_sheet_functions if user_defined_sheet_functions is not None else [])]
    return user_defined_functions
    

def validate_user_defined_editors(user_defined_editors: Optional[List[Callable]]) -> List[Callable]:
    if user_defined_editors is None:
        return []

    for f in user_defined_editors:
        df_arguments = []
        parameters = inspect.signature(f).parameters
        has_previous_dataframe_arg = False
        for param_name in parameters:
            annotation = parameters[param_name].annotation
            if annotation == pd.DataFrame:
                df_arguments.append(param_name)
                has_previous_dataframe_arg = True

            if annotation == ColumnHeader:
                if not has_previous_dataframe_arg:
                    raise ValueError(f"Editor {f.__name__} has a ColumnHeader argument {param_name}, but does not have a pd.DataFrame argument preceeding it. Please ensure that the first argument is a pd.DataFrame.")

        if len(df_arguments) != 1:
            raise ValueError(f"Editor {f.__name__} must only have a single dataframe parameter, but instead got {df_arguments} as arguments")
        
    return user_defined_editors


def get_user_defined_function_param_type(f: Callable, param_name: str) -> UserDefinedFunctionParamType:

    parameters = inspect.signature(f).parameters
    param_type = parameters[param_name].annotation

    if param_type == str:
        return 'str'
    elif param_type == int:
        return 'int'
    elif param_type == float:
        return 'float'
    elif param_type == bool:
        return 'bool'
    elif param_type == pd.DataFrame:
        return 'DataFrame'
    elif param_type == ColumnHeader:
        return 'ColumnHeader'
    else:
        return 'any'


def get_param_names_to_types_for_user_defined_functon(f: Callable) -> Dict[str, UserDefinedFunctionParamType]:
    param_names_to_types = {}

    for name in inspect.signature(f).parameters:
        param_names_to_types[name] = get_user_defined_function_param_type(f, name)

    return param_names_to_types

def get_domain_from_user_defined_function(f: Callable) -> Optional[str]:
    """
    Given a user defined importer, return the domain that it's associated with. If it's not associated with a domain, return None.

    Currently, we just support starting a doc string line with "Domain: " to specify the domain.
    TODO: make this more robust in the future
    """
    
    docstring = f.__doc__
    if docstring is None:
        return None

    for line in docstring.lower().split('\n'):
        if line.strip().startswith('domain: '):
            return line.strip().split('domain: ')[1].strip()
    
    return None

def get_docstring_for_user_defined_function(f: Callable) -> Optional[str]:
    """
    Given a user defined importer, return the docstring for it. Will cleanup extra
    whitespace due to the way that docstrings are formatted in Python.

    Will also get rid of the domain line if it exists.
    """

    docstring = f.__doc__
    if docstring is None:
        return None
    
    lines = docstring.split("\n")
    # Remove the domain line if it exists
    lines = [line for line in lines if not line.strip().lower().startswith('domain: ')]
    return ' '.join(lines)

def get_user_defined_importers_for_frontend(state: Optional[State]) -> List[Any]:
    if state is None:
        return []

    return [
        {
            'name': f.__name__,
            'docstring': get_docstring_for_user_defined_function(f),
            'parameters': get_param_names_to_types_for_user_defined_functon(f),
            'domain': get_domain_from_user_defined_function(f)
        }
        for f in state.user_defined_importers
    ]

def get_user_defined_editors_for_frontend(state: Optional[State]) -> List[Any]:
    if state is None:
        return []

    return [
        {
            'name': f.__name__,
            'docstring': f.__doc__,
            'parameters': get_param_names_to_types_for_user_defined_functon(f),
        }
        for f in state.user_defined_editors
    ]

def get_importer_params_and_type_and_value(f: Callable, frontend_params: Dict[str, str]) -> Dict[str, Tuple[UserDefinedFunctionParamType, str]]:
    importer_params_and_type_and_value = {}

    for param_name, param_value in frontend_params.items():
        param_type = get_user_defined_function_param_type(f, param_name)
        importer_params_and_type_and_value[param_name] = (param_type, param_value)
    
    return importer_params_and_type_and_value


def get_user_defined_function_param_type_and_execute_value_and_transpile_value(
        state: State,
        f: Callable, 
        frontend_params: Dict[str, str]
    ) -> Dict[str, Tuple[UserDefinedFunctionParamType, Any, Any]]: # TODO: document this type better
    """
    Given a current state, a user defined function f, and the frontend params that the user configured
    for this function f, return a dictionary mapping the parameter name to a tuple:
    1.  parameter type
    2.  the value to be passed to the function when executing it in the step performer
    3.  the value to be transpiled in the code chunk for the parameter

    This function is responsible for casting the frontend params to the appropriate type, and for
    transpiling the values to be passed to the function. It throws appropriate errors if the frontend
    params cannot be cast to the appropriate type.
    """
    user_defined_function_params: Dict[str, Tuple[UserDefinedFunctionParamType, Any, Any]] = {}

    sheet_index = None
    for param_name, (param_type, param_value) in get_importer_params_and_type_and_value(f, frontend_params).items():
        try:
            if param_type == 'DataFrame':
                sheet_index = state.df_names.index(param_value)
                df = state.dfs[sheet_index]
                # Because we want to just transpile the dataframe name, the third tuple value (the value to be transpiled) should
                # not be wrapped in get_column_header_as_transpiled_code
                # NOTE: we also make a copy of the DF to avoid issues with it being modified by the calling function
                user_defined_function_params[param_name] = (param_type, df.copy(), param_value)
            elif param_type == 'ColumnHeader':
                # If it's a column header, then we should have already gotten the sheet index
                if sheet_index is None:
                    raise ValueError(f"Parameter {param_name} is a ColumnHeader, but is not directly preceeded by a pd.DataFrame argument. As such, it's referencing an ambigious dataframe.")
                
                # notably, the column header type will actually be provided as a column id? Does that make sense? I think so -- that's what we can 
                # select on in the frontend. TODO: think about if this makes sense with the way replaying works...
                column_header = state.column_ids.get_column_header_by_id(sheet_index, param_value)
                user_defined_function_params[param_name] = (param_type, column_header, get_column_header_as_transpiled_code(param_value))
            elif param_type == 'str':
                user_defined_function_params[param_name] = (param_type, param_value, get_column_header_as_transpiled_code(param_value))
            elif param_type == 'int':
                execute_value: Any = int(param_value)
                user_defined_function_params[param_name] = (param_type, execute_value, get_column_header_as_transpiled_code(execute_value))
            elif param_type == 'float':
                execute_value = float(param_value) 
                user_defined_function_params[param_name] = (param_type, execute_value, get_column_header_as_transpiled_code(execute_value))
            elif param_type == 'bool':
                execute_value = 'true' in param_value.lower()
                user_defined_function_params[param_name] = (param_type, execute_value, get_column_header_as_transpiled_code(execute_value))
            else:
                try:
                    # If we don't know the type (it's untyped), we try and convert the value to a python object -- using an eval. If that fails, 
                    # we just keep it as a string
                    execute_value = eval(param_value)
                    user_defined_function_params[param_name] = (param_type, execute_value, get_column_header_as_transpiled_code(execute_value))
                except:
                    # If we cannot eval the result, it's likely a string, so we just pass it through
                    user_defined_function_params[param_name] = (param_type, param_value, get_column_header_as_transpiled_code(param_value))
        except:
            raise MitoError(
                'user_defined_function_error',
                f"Function {f.__name__} raised an error.",
                f"Parameter {param_name} with value {param_value} cannot be cast to type {param_type}. Please insert an appropriate value.",
                error_modal=False
            )

    return user_defined_function_params

def get_transpiled_user_defined_function_params(user_defined_function_param_types_and_values: Dict[str, Tuple[UserDefinedFunctionParamType, Any, Any]]) -> str:
    param_strings = []
    for param_name, (_, _, transpiled_value) in user_defined_function_param_types_and_values.items():
        param_strings.append(f'{param_name}={transpiled_value}')
    return ", ".join(param_strings)



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict, List, Optional, Set, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.sort_code_chunk import SortCodeChunk
from mitosheet.errors import make_invalid_sort_error
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param
from mitosheet.types import ColumnID

# CONSTANTS USED IN THE SORT STEP ITSELF
SORT_DIRECTION_ASCENDING = 'ascending'
SORT_DIRECTION_DESCENDING = 'descending'
SORT_DIRECTION_NONE = 'none'

class SortStepPerformer(StepPerformer):
    """
    Allows you to sort a df based on key column, in either
    ascending or descending order.
    """

    @classmethod
    def step_version(cls) -> int:
        return 2

    @classmethod
    def step_type(cls) -> str:
        return 'sort'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        try: 
            return cls.execute_through_transpile(
                prev_state, 
                params, 
            )
        except TypeError as e:
            # A TypeError occurs when you try to sort a column with incomparable 
            # dtypes (ie: a column with strings and floats)
            
            sheet_index: int = get_param(params, 'sheet_index')
            column_id: ColumnID = get_param(params, 'column_id')
            column_header = prev_state.column_ids.get_column_header_by_id(sheet_index, column_id)

            raise make_invalid_sort_error(column_header)


    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            SortCodeChunk(
                prev_state, 
                get_param(params, 'sheet_index'),
                get_param(params, 'column_id'),
                get_param(params, 'sort_direction')
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import json
from typing import Any, Dict, List, Tuple
from mitosheet.code_chunks.code_chunk_utils import get_code_chunks
from mitosheet.types import ParamSubtype, ParamType, ParamName, ParamValue, StepsManagerType, ParamMetadata
from mitosheet.updates.args_update import is_string_arg_to_mitosheet_call
from mitosheet.transpiler.transpile_utils import get_final_function_params_with_subtypes_turned_to_parameters

def get_parameterizable_params(params: Dict[str, Any], steps_manager: StepsManagerType) -> List[Tuple[ParamValue, ParamType, ParamSubtype]]:

        all_parameterizable_params: List[Tuple[ParamValue, ParamType, ParamSubtype]] = []

        # First, get the original arguments to the mitosheet - we only let you parameterize df names for now
        for arg in steps_manager.original_args_raw_strings:
                if not is_string_arg_to_mitosheet_call(arg):
                        all_parameterizable_params.append((arg, 'import', "import_dataframe")) # type: ignore
    
        # Get optimized code chunk, and get their parameterizable params
        code_chunks = get_code_chunks(steps_manager.steps_including_skipped[:steps_manager.curr_step_idx + 1], optimize=True)

        for code_chunk in code_chunks:
                parameterizable_params = code_chunk.get_parameterizable_params()
                all_parameterizable_params.extend(parameterizable_params)

        return all_parameterizable_params

# This function is used to get the parameterizable params for the UI in streamlit. 
# It returns a list of ParamMetadata objects, which are defined in mitosheet/types.py
# The ParamMetadata object includes various information that a streamlit developer might
# need to display the parameterizable params in the UI.
def get_parameterizable_params_metadata(steps_manager: StepsManagerType) -> List[ParamMetadata]:
        # First, get all the parameterizable params and the function params
        all_parameterizable_params = get_parameterizable_params({}, steps_manager)
        function_params: Dict[ParamName, ParamValue] = get_final_function_params_with_subtypes_turned_to_parameters(steps_manager, 'all')
        param_names = list(function_params.keys())

        # Now, we need to map the parameterizable params to the ParamMetadata type. 
        def map_tuple_to_param_data(args):
                param = args[0]
                param_name = args[1] 
                return {
                        # Removes the quotes and r-string from the param value because they aren't needed for streamlit. 
                        'original_value': param[0][2:-1] if param[0][:2] == "r'" else param[0],
                        'type': param[1],
                        'subtype': param[2],
                        # If the param is a df_name, then it is required for the function
                        # because we won't be able to store the initial value in the RunnableAnalysis for streamlit.
                        'required': param[2] == 'import_dataframe',
                        'name': param_name
                }

        return list(map(map_tuple_to_param_data, zip(all_parameterizable_params, param_names)))


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

"""
This file contains utilities for the rest of the mitosheet package
to interact with telemetry. See the README.md file in this folder for
more details.
"""

import datetime
import os
import platform
import sys
import time
from copy import copy
from typing import Any, Dict, Optional

import requests

from unittest.mock import patch
from mitosheet.enterprise.telemetry.mito_log_uploader import MitoLogUploader
from mitosheet.errors import MitoError, get_recent_traceback_as_list
from mitosheet.telemetry.anonymization_utils import anonymize_object, get_final_private_params_for_single_kv
from mitosheet.telemetry.private_params_map import LOG_EXECUTION_DATA_LENGTH_FIRST_ELEMENT, LOG_EXECUTION_DATA_PUBLIC
from mitosheet.types import StepsManagerType
from mitosheet.user.location import get_location, is_docker, is_jupyterlite
from mitosheet.user.schemas import UJ_FEEDBACKS, UJ_FEEDBACKS_V2, UJ_INTENDED_BEHAVIOR, UJ_MITOSHEET_TELEMETRY, UJ_USER_EMAIL
from mitosheet.user.utils import is_local_deployment, is_pro

WRITE_KEY = '6I7ptc5wcIGC4WZ0N1t0NXvvAbjRGUgX' 

import analytics
analytics.write_key = WRITE_KEY

if is_jupyterlite():
    # If we are in JupyterLite, we have to do a few things to get telemetry working:
    # 1. We have to set the sync_mode to True, so that we don't start a thread
    # 2. We have to patch the requests library to use pyodide's fetch instead of requests, which
    #    does not work in JupyterLite. We do this with the pyodide_http library
    # 3. When we call the identify, we actuall have to mock the Session.post function, 
    #    as pyodide_http doesn't do this (it just fixes requests.post). We do this at the
    #    call site with unittest.mock.patch

    analytics.sync_mode = True

    import pyodide_http
    pyodide_http.patch_all()

    # Wrapper
    def post(*args, **kwargs):
        return requests.post(args[1], **kwargs)


from mitosheet._version import __version__, package_name
from mitosheet.errors import MitoError, get_recent_traceback_as_list
from mitosheet.user import (UJ_STATIC_USER_ID, get_user_field,
                            is_local_deployment, is_running_test)

# If you want, you can optionally choose to print logs
PRINT_LOGS = False


try:
    import mitosheet_helper_private
    MITOSHEET_HELPER_PRIVATE = True
except ImportError:
    MITOSHEET_HELPER_PRIVATE = False


def telemetry_turned_on() -> bool:
    """
    Helper function that tells you if logging is turned on or
    turned off on the entire Mito instance
    """
    # If private helper is installed, then we don't log anything
    if MITOSHEET_HELPER_PRIVATE:
        return False

    # Check if the config is set
    if os.environ.get('MITO_CONFIG_FEATURE_TELEMETRY') is not None:
        from mitosheet.enterprise.mito_config import is_env_variable_set_to_true
        return is_env_variable_set_to_true(os.environ.get('MITO_CONFIG_FEATURE_TELEMETRY', ''))

    # If Mito Pro is on, then don't log anything
    if is_pro():
        return False

    telemetry = get_user_field(UJ_MITOSHEET_TELEMETRY) 
    return telemetry if telemetry is not None else False

def _get_anonymized_log_params(params: Dict[str, Any], steps_manager: Optional[StepsManagerType]=None) -> Dict[str, Any]:
    """
    Private params are where we _make sure_ that no private
    user data leaves the user's machine. We replace any potentially
    non-private params with private versions of them here.
    """
    private_params: Dict[str, Any] = dict()

    for key, value in params.items():
        private_params = {
            **private_params, 
            **get_final_private_params_for_single_kv(key, value, params, steps_manager)
        }

    # Prefix all the params with params_ so we can easily find them
    private_params = {'params_' + key: value for key, value in private_params.items()}
            
    return private_params


def _get_execution_data_log_params(steps_manager: Optional[StepsManagerType]=None) -> Dict[str, Any]:
    """
    Get the execution params as well, again making sure
    to remove any private data.
    """
    execution_data_params = {}

    # First, try and get the execution data from the stpe
    if steps_manager and steps_manager.curr_step.execution_data:
        for key, value in steps_manager.curr_step.execution_data.items():
            # Only take those items that are marked as public
            if key in LOG_EXECUTION_DATA_PUBLIC:
                execution_data_params['execution_data_' + key] = value
            elif key in LOG_EXECUTION_DATA_LENGTH_FIRST_ELEMENT:
                # Calculate the length, if we're asked to
                execution_data_params['execution_data_' + key] = len(value[0]) if value else 0
            # And make the rest private
            else:
                execution_data_params['execution_data_' + key] = anonymize_object(value)

    return execution_data_params

def _get_wsc_log_params(steps_manager: Optional[StepsManagerType]=None) -> Dict[str, Any]:
    """
    Get data from the widget state container that is useful for any
    log event. Note that none of this is private data.
    """

    if steps_manager:
        # We also get some metadata about the widget state container at this state
        # NOTE: we keep this as underscored with wsc for backwards compatibility with
        # our logs!
        return {
            'wsc_analysis_name': steps_manager.analysis_name,
            # NOTE: Change this when code fixing this logic is merged in
            'wsc_local': is_local_deployment(),
            'wsc_curr_step_idx': steps_manager.curr_step_idx,
            'wsc_curr_step_type': steps_manager.curr_step.step_type,
            'wsc_public_interface_version': steps_manager.public_interface_version,
        }
    else:
        return {}

def _get_error_log_params(failed: bool=False, mito_error: Optional[MitoError]=None)-> Dict[str, Any]:
    """
    Get relevant logging data from any recently thrown error
    """

    # We also check there is an edit_error, and if there is, then we add the error logs
    if mito_error is not None or failed:
        recent_traceback = get_recent_traceback_as_list() 
        # TODO: if this is to long, we should chop it, and also take only the first 10k characters of the last line...
        return {
            'error_traceback': recent_traceback,
            # We get the last line of the error as it makes it much easier
            # to easily analyze on error messages 
            'error_traceback_last_line': recent_traceback[-1],
        }
    else:
        return {}

def _get_processing_time_log_params(steps_manager: Optional[StepsManagerType]=None, start_time: Optional[float]=None)-> Dict[str, Any]:
    """
    Get data relevant for measuring performance impact
    """

    processing_time_params = {}
    # We also log some timing information - which we round to a single decimal place just
    # so that we can bucket these items easily. Note we include a variety of roundings of 
    # the time, so that we can make sure to aggregate in Mixpanel well (which will die if 
    # it is given to many values).
    if start_time is not None:
        processing_time = time.perf_counter() - start_time
        processing_time_params['processing_time'] = round(processing_time, 1)
        processing_time_params['processing_time_seconds'] = int(round(processing_time, 0))
        processing_time_params['processing_time_seconds_ten'] = int(round(processing_time, -1))
        processing_time_params['processing_time_seconds_hundred'] = int(round(processing_time, -2))

        # If we just did an update, and this update has a pandas processing time, then we can calculate the
        # time that we spent as mito overhead vs. just executing pandas code
        if steps_manager and steps_manager.curr_step.execution_data and 'pandas_processing_time' in steps_manager.curr_step.execution_data:
            pandas_processing_time = steps_manager.curr_step.execution_data['pandas_processing_time']
            processing_time_params['processing_time_pandas'] = round(pandas_processing_time, 1)
            processing_time_params['processing_time_pandas_seconds'] = int(round(pandas_processing_time, 0))
            processing_time_params['processing_time_pandas_seconds_ten'] = int(round(pandas_processing_time, -1))
            processing_time_params['processing_time_pandas_seconds_hundred'] = int(round(pandas_processing_time, -2))

            # And we explicitly calculate the overhead
            overhead_processing_time = round(processing_time - pandas_processing_time, 1)
            processing_time_params['processing_time_overhead'] = overhead_processing_time
            processing_time_params['processing_time_overhead_seconds'] = int(round(overhead_processing_time, 0))
            processing_time_params['processing_time_overhead_seconds_ten'] = int(round(overhead_processing_time, -1))
            processing_time_params['processing_time_overhead_seconds_hundred'] = int(round(overhead_processing_time, -2))
        
    return processing_time_params


try:
    from jupyterlab import __version__ as jupyterlab_version
except:
    jupyterlab_version = 'No JupyterLab'
try:
    from notebook import __version__ as notebook_version
except:
    notebook_version = 'No notebook'
try:
    from pandas import __version__ as pandas_version
except:
    pandas_version = 'no pandas'
try:
    from ipywidgets import __version__ as ipywidgets_version
except:
    ipywidgets_version = 'no pandas'
try:
    # Format version_python as "3.8.5"
    version_python = '.'.join([str(x) for x in sys.version_info[:3]])
except:
    version_python = 'no python'


__location = None

def _get_environment_params() -> Dict[str, Any]:
    """
    Get data relevant for tracking the environment, so we can 
    ensure Mitosheet compatibility with any system
    """
    global __location
    if __location is None:
        __location = get_location()
    
    # Add the python properties to every log event we can
    environment_params = {
        'version_python': version_python,
        'version_pandas': pandas_version,
        'version_ipywidgets': ipywidgets_version,
        'version_jupyterlab': jupyterlab_version,
        'version_notebook': notebook_version,
        'version_mito': __version__,
        'package_name': package_name,
        'location': __location,
        'is_docker': is_docker()
    }

    return environment_params

experiment = None
def _get_experiment_params() -> Dict[str, Any]:
    """
    Get data relevant for tracking the experiment, so we can 
    see how the experiment is running

    NOTE: This must match the function in the installer
    """
    global experiment
    if experiment is None:
        from mitosheet.experiments.experiment_utils import get_current_experiment
        experiment = get_current_experiment()

    if experiment is None:
        experiment_params = {'experiment_id': 'No experiment'}
    else:
        experiment_params = {
            'experiment_id': experiment["experiment_id"],
            'experiment_variant': experiment["variant"],
            f'experiment_{experiment["experiment_id"]}': experiment['variant']
        }

    return experiment_params

def log_event_processed(event: Dict[str, Any], steps_manager: StepsManagerType, failed: bool=False, mito_error: Optional[MitoError]=None, start_time: Optional[float]=None) -> None:
    """
    Helper function for logging when an event is processed by the backend,
    including an edit event, an api call, or an update event. 

    It generates mulitple logs so that aggregating and breaking down this 
    data is easier.

    NOTE: if processing the event fails, then failed should be True. If there was an
    edit error that was thrown during the processing of the event, then edit_error
    should be set to that error.
    """

    # We choose to log the event type, as it is the best high-level item for our logs
    # and we append a _failed if the event failed in doing this.
    log_event: str = event['type'] + ('_failed' if failed else '')
    params_for_final_log = {
        # NOTE: We make a copy here so we don't modify the actual params
        # dict, which we don't want to do as it's used elsewhere!
        'params': copy(event['params']),
        'steps_manager': steps_manager,
        'failed': failed,
        'mito_error': mito_error,
        'start_time': start_time
    }

    log(
        log_event, 
        **params_for_final_log
    )

    if failed:
        # We also generate a double log in the case of errors, whenever anything fails. This allows
        # us to easily track the number of users who are getting errors
        log(
            'error', 
            **params_for_final_log
        )
    else:
        # We also generate a single aggregate log for each of the different
        # types of events the user sends to the backend. As with the above, 
        # this allows us to easily aggregate across different types of events
        # and track general trends (at the cost of making the logs look a bit
        # messier to human eyes)
        params_for_final_log['params']['log_event'] = log_event
        log(
            event['event'], 
            **params_for_final_log
        )



def identify() -> None:
    """
    Helper function for identifying a user. We just take
    their python version, mito version, and email.
    """
    if not telemetry_turned_on():
        return

    static_user_id = get_user_field(UJ_STATIC_USER_ID)
    user_email = get_user_field(UJ_USER_EMAIL)
    intended_behavior = get_user_field(UJ_INTENDED_BEHAVIOR)
    feedbacks = get_user_field(UJ_FEEDBACKS)
    feedbacks_v2 = get_user_field(UJ_FEEDBACKS_V2)
    local = is_local_deployment()
    jupyterlite = is_jupyterlite()
    operating_system = platform.system()

    params = {
        'version_python': version_python,
        'version_pandas': pandas_version,
        'version_ipywidgets': ipywidgets_version,
        'version_sys': sys.version,
        'version_mito': __version__,
        'package_name': package_name, 
        'version_jupyterlab': jupyterlab_version,
        'version_notebook': notebook_version,
        'operating_system': operating_system,
        'email': user_email,
        'local': local,
        'jupyterlite': jupyterlite,
        UJ_INTENDED_BEHAVIOR: intended_behavior,
        UJ_FEEDBACKS: feedbacks,
        UJ_FEEDBACKS_V2: feedbacks_v2
    }

    if not is_running_test():

        if is_jupyterlite():
            # We patch the post function to use pyodide fetch
            # instead of requests
            with patch('requests.sessions.Session.post', post):
                analytics.identify(static_user_id, params)

        else:        
            analytics.identify(static_user_id, params)


def log(
        log_event: str, 
        params: Optional[Dict[str, Any]]=None, 
        steps_manager: Optional[StepsManagerType]=None, 
        failed: bool=False, 
        mito_error: Optional[MitoError]=None, 
        start_time: Optional[float]=None,
    ) -> None:
    """
    This function is the entry point for all logging. It collects
    all relevant parameters, exeuction data, and more info while
    making sure to anonymize all data. 

    Then, if telemetry is not turned off and we are not running tests,
    we log this information.
    """
    if params is None:
        params = {}


    final_params: Dict[str, Any] = {}

    # First, get the private params
    final_params = {**final_params, **_get_anonymized_log_params(params, steps_manager=steps_manager)}

    # Then, get the execution data from the steps
    final_params = {**final_params, **_get_execution_data_log_params(steps_manager=steps_manager)}
    
    # Then, get the logs for the specific analysis we're in
    final_params = {**final_params, **_get_wsc_log_params(steps_manager=steps_manager)}

    # Then, get the logs for the error (if there is one)
    final_params = {**final_params, **_get_error_log_params(failed=failed, mito_error=mito_error)}

    # Then, get the logs for the processing time of the operation
    final_params = {**final_params, **_get_processing_time_log_params(steps_manager=steps_manager, start_time=start_time)}

    # Then, get the params for the environment 
    final_params = {**final_params, **_get_environment_params()}

    # Then, get the params for the all experiments
    final_params = {**final_params, **_get_experiment_params()}

    # Then, make sure to add the user email
    final_params['email'] = get_user_field(UJ_USER_EMAIL)

    # Finially, do the acutal logging. We do not log anything when tests are
    # running, or if telemetry is turned off
    if not is_running_test() and telemetry_turned_on():

        if is_jupyterlite():
            # We patch post function to use pyodide fetch
            # instead of requests
            with patch('requests.sessions.Session.post', post):
                analytics.track(
                    get_user_field(UJ_STATIC_USER_ID), 
                    log_event, 
                    final_params
                )
        else:
            analytics.track(
                get_user_field(UJ_STATIC_USER_ID), 
                log_event, 
                final_params
            )
        

    # If we want to print the logs for debugging reasons, then we print them as well
    if PRINT_LOGS:
        print(
            log_event, 
            final_params
        )

    analytics_url = steps_manager.mito_config.analytics_url if steps_manager is not None else None
    if analytics_url is not None:
        requests.post(
            analytics_url,
            json={
                'user_id': get_user_field(UJ_STATIC_USER_ID),
                'log_event': log_event
            }
        )

    mito_log_uploader = steps_manager.mito_log_uploader if steps_manager is not None else None
    if mito_log_uploader is not None:
        mito_log_uploader.log(log_event, final_params)




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import List, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import get_column_header_as_transpiled_code
from mitosheet.types import ColumnID

class OneHotEncodingCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, sheet_index: int, column_id: ColumnID):
        super().__init__(prev_state)
        self.sheet_index = sheet_index
        self.column_id = column_id

        self.column_header = self.prev_state.column_ids.get_column_header_by_id(self.sheet_index, self.column_id)

    def get_display_name(self) -> str:
        return 'One Hot Encoding'
    
    def get_description_comment(self) -> str:
        return f'One-hot Encoded {self.column_header}'
        
    def get_code(self) -> Tuple[List[str], List[str]]:
        df_name = self.prev_state.df_names[self.sheet_index]
        column_index = self.prev_state.dfs[self.sheet_index].columns.tolist().index(self.column_header)

        transpiled_column_header = get_column_header_as_transpiled_code(self.column_header)

        encode_line = f'tmp_df = pd.get_dummies({df_name}[{transpiled_column_header}])'
        add_line = f'{df_name}[tmp_df.columns] = tmp_df'
        reorder_columns_line = f'{df_name} = {df_name}[{df_name}.columns[:{column_index + 1}].tolist() + tmp_df.columns.tolist() + {df_name}.columns[{column_index + 1}:-len(tmp_df.columns)].tolist()]'

        return [
            encode_line,
            add_line,
            reorder_columns_line   
        ], ['import pandas as pd']

    
    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains all functions that can be used in a sheet that operate on
numbers.

All functions describe their behavior with a function documentation object
in the function docstring. Function documentation objects are described
in more detail in docs/README.md.

NOTE: This file is alphabetical order!
"""
import decimal
import math
import sys
from datetime import datetime
from typing import Optional, Union

import numpy as np
import pandas as pd

from mitosheet.errors import MitoError
from mitosheet.is_type_utils import is_datetime_dtype
from mitosheet.public.v3.errors import handle_sheet_function_errors
from mitosheet.public.v3.rolling_range import RollingRange
from mitosheet.public.v3.sheet_functions.utils import (
    get_final_result_series_or_primitive, get_index_from_series,
    get_series_from_primitive_or_series)
from mitosheet.public.v3.types.decorators import (
    cast_values_in_all_args_to_type, cast_values_in_arg_to_type)
from mitosheet.public.v3.types.sheet_function_types import (
    DatetimeFunctionReturnType, DatetimeRestrictedInputType,
    FloatFunctonReturnType, IntFunctionReturnType, IntRestrictedInputType,
    NumberFunctionReturnType, NumberInputType, NumberRestrictedInputType)


@cast_values_in_arg_to_type('arg', 'number')
@handle_sheet_function_errors
def ABS(arg: NumberRestrictedInputType) -> NumberFunctionReturnType:
    """
    {
        "function": "ABS",
        "description": "Returns the absolute value of the passed number or series.",
        "search_terms": ["abs", "absolute value"],
        "category": "MATH",
        "examples": [
            "ABS(-1.3)",
            "ABS(A)"
        ],
        "syntax": "ABS(value)",
        "syntax_elements": [{
                "element": "value",
                "description": "The value or series to take the absolute value of."
            }
        ]
    }
    """
    if isinstance(arg, int) or isinstance(arg, float):
        return abs(arg)

    return arg.abs() # type: ignore


@cast_values_in_all_args_to_type('number')
@handle_sheet_function_errors
def AVG(*argv: Optional[NumberInputType]) -> NumberFunctionReturnType:
    """
    {
        "function": "AVG",
        "description": "Returns the numerical mean value of the passed numbers and series.",
        "search_terms": ["avg", "average", "mean"],
        "category": "MATH",
        "examples": [
            "AVG(1, 2)",
            "AVG(A, B)",
            "AVG(A, 2)"
        ],
        "syntax": "AVG(value1, [value2, ...])",
        "syntax_elements": [{
                "element": "value1",
                "description": "The first number or series to consider when calculating the average."
            },
            {
                "element": "value2, ... [OPTIONAL]",
                "description": "Additional numbers or series to consider when calculating the average."
            }
        ]
    }
    """

    # Calculate the sum using the SUM function
    sum_for_avg = SUM(*argv)

    # Then, we count the number of entries in the arguements. To match Excel's behavior, we do not count 
    # nan values, and we do not count when a RollingRange runs off the end of the sum
    num_entries: Union[pd.Series,int] = 0

    for arg in argv:
        if isinstance(arg, pd.DataFrame):
            num_non_null_values = arg.count().sum()
            num_entries += int(num_non_null_values)

        elif isinstance(arg, RollingRange):
            num_non_null_values_series = arg.apply(lambda df: df.count().sum())
            num_entries += num_non_null_values_series
            
        elif isinstance(arg, pd.Series):
            # Because series are summed row-wise, we need to count the number of non-null
            # values in the row, which we do with this little type cast
            num_entries += (~arg.isna()).astype(int)
            
        elif arg is not None:
            num_entries += 1

    there_are_entries = (isinstance(num_entries, int) and num_entries != 0) or isinstance(num_entries, pd.Series)

    return sum_for_avg / num_entries if there_are_entries else 0

@cast_values_in_arg_to_type('s1', 'number')
@cast_values_in_arg_to_type('s2', 'number')
@handle_sheet_function_errors
def CORR(s1: NumberRestrictedInputType, s2: NumberRestrictedInputType) -> NumberFunctionReturnType:
    """
    {
        "function": "CORR",
        "description": "Computes the correlation between two series, excluding missing values.",
        "search_terms": ["corr", "correlation", "r^2"],
        "category": "MATH",
        "examples": [
            "=CORR(A, B)",
            "=CORR(B, A)"
        ],
        "syntax": "CORR(series_one, series_two)",
        "syntax_elements": [{
                "element": "series_one",
                "description": "The number series to convert to calculate the correlation."
            }, {
                "element": "series_two",
                "description": "The number series to convert to calculate the correlation."
            }
        ]
    }
    """
    if isinstance(s1, int) or isinstance(s1, float) or isinstance(s2, int) or isinstance(s2, float):
        return 0
    
    return s1.corr(s2, method='pearson') # type: ignore
 

@cast_values_in_arg_to_type('arg', 'number')
@handle_sheet_function_errors
def EXP(arg: NumberRestrictedInputType) -> NumberFunctionReturnType:
    """
    {
        "function": "EXP",
        "description": "Returns e, the base of the natural logarithm, raised to the power of passed series.",
        "search_terms": ["exp", "exponent", "log", "natural log"],
        "category": "MATH",
        "examples": [
            "=EXP(data)",
            "=EXP(A)"
        ],
        "syntax": "EXP(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to raise e to."
            }
        ]
    }
    """
    if isinstance(arg, int) or isinstance(arg, float):
        return math.exp(arg)
    
    return pd.Series(np.exp(arg))


@cast_values_in_arg_to_type('arg', 'float')
@handle_sheet_function_errors
def FLOAT(arg: NumberRestrictedInputType) -> FloatFunctonReturnType:
    """
    {
        "function": "FLOAT",
        "description": "Converts a string series to a float series. Any values that fail to convert will return NaN.",
        "search_terms": ["number", "to number"],
        "category": "MATH",
        "examples": [
            "=FLOAT(Prices_string)",
            "=FLOAT('123.123')"
        ],
        "syntax": "FLOAT(string_series)",
        "syntax_elements": [{
                "element": "string_series",
                "description": "The series or string to convert to a float."
            }
        ]
    }
    """
    if isinstance(arg, int) or isinstance(arg, float):
        return float(arg)
    
    return arg.fillna(np.nan) # type: ignore

@cast_values_in_arg_to_type('arg', 'int')
@handle_sheet_function_errors
def INT(arg: IntRestrictedInputType) -> IntFunctionReturnType:
    """
    {
        "function": "INT",
        "description": "Converts a string series to a int series. Any values that fail to convert will return 0.",
        "search_terms": ["number", "to integer"],
        "category": "MATH",
        "examples": [
            "=INT(Prices_string)",
            "=INT('123')"
        ],
        "syntax": "INT(string_series)",
        "syntax_elements": [{
                "element": "string_series",
                "description": "The series or string to convert to a int."
            }
        ]
    }
    """
    return arg
    


@cast_values_in_arg_to_type('arg', 'number')
@handle_sheet_function_errors
def KURT(arg: NumberRestrictedInputType) -> NumberFunctionReturnType:
    """
    {
        "function": "KURT",
        "description": "Computes the unbiased kurtosis, a measure of tailedness, of a series, excluding missing values.",
        "search_terms": ["kurtosis"],
        "category": "MATH",
        "examples": [
            "=KURT(A)",
            "=KURT(A * B)"
        ],
        "syntax": "KURT(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to calculate the unbiased kurtosis of."
            }
        ]
    }
    """
    if isinstance(arg, int) or isinstance(arg, float):
        return 0
        
    return arg.kurt() # type: ignore


@cast_values_in_arg_to_type('arg', 'number')
@cast_values_in_arg_to_type('base', 'number')
@handle_sheet_function_errors
def LOG(arg: NumberRestrictedInputType, base: Optional[NumberRestrictedInputType]=None) -> NumberFunctionReturnType:
    """
    {
        "function": "LOG",
        "description": "Calculates the logarithm of the passed series with an optional base.",
        "search_terms": ["log", "logarithm", "natural log"],
        "category": "MATH",
        "examples": [
            "LOG(10) = 1",
            "LOG(100, 10) = 2"
        ],
        "syntax": "LOG(series, [base])",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to take the logarithm of."
            },
            {
                "element": "base [OPTIONAL]",
                "description": "The base of the logarithm to use. Defaults to 10 if no base is passed."
            }
        ]
    }
    """
    
    if base is None:
        base = 10

    if (isinstance(arg, int) or isinstance(arg, float)) and (isinstance(base, int) or isinstance(base, float)):
        return math.log(arg, base)
    
    index = get_index_from_series(arg, base)
    arg = get_series_from_primitive_or_series(arg, index)
    base = get_series_from_primitive_or_series(base, index).fillna(10)
    
    # See here: https://stackoverflow.com/questions/25169297/numpy-logarithm-with-base-n
    return pd.Series(np.log(arg) / np.log(base)) # type: ignore

@cast_values_in_all_args_to_type('number', ['datetime'])
@handle_sheet_function_errors
def MAX(*argv: Union[NumberInputType, None, DatetimeRestrictedInputType]) -> Union[NumberFunctionReturnType, DatetimeFunctionReturnType]:
    """
    {
        "function": "MAX",
        "description": "Returns the maximum value among the passed arguments.",
        "search_terms": ["max", "maximum", "minimum"],
        "category": "MATH",
        "examples": [
            "MAX(10, 11)",
            "MAX(Old_Data, New_Data)"
        ],
        "syntax": "MAX(value1, [value2, ...])",
        "syntax_elements": [{
                "element": "value1",
                "description": "The first number or column to consider for the maximum value."
            },
            {
                "element": "value2, ... [OPTIONAL]",
                "description": "Additional numbers or columns to compute the maximum value from."
            }
        ]
    }
    """

    # If the user passes numbers, we start the default as the smallest possible number
    # but if it's a datetime, we start the default as the earliest possible date
    default_value = -sys.maxsize - 1
    for arg in argv:
        if isinstance(arg, pd.Series) and is_datetime_dtype(str(arg.dtype)):
            default_value = pd.Timestamp.min
            break
        elif isinstance(arg, pd.Timestamp) or isinstance(arg, datetime):
            default_value = pd.Timestamp.min
            break
        elif isinstance(arg, pd.DataFrame):
            # Check if any series are datetimes
            for col in arg.columns:
                if is_datetime_dtype(str(arg[col].dtype)):
                    default_value = pd.Timestamp.min
                    break
        elif isinstance(arg, RollingRange):
            # Check if any series are datetimes
            for col in arg.obj.columns:
                if is_datetime_dtype(str(arg.obj[col].dtype)):
                    default_value = pd.Timestamp.min
                    break

    result = get_final_result_series_or_primitive(
        default_value,
        argv,
        lambda df: df.max().max(),
        lambda previous_value, new_value: max(previous_value, new_value),
        lambda previous_series, new_series: pd.concat([previous_series, new_series], axis=1).max(axis=1)
    )

    # If we don't find any arguements, we default to 0 -- like Excel -- even for numbers
    kept_default_max_value = not isinstance(result, pd.Series) and result == (default_value)
    return result if not kept_default_max_value else 0


@cast_values_in_all_args_to_type('number', ['datetime'])
@handle_sheet_function_errors
def MIN(*argv: Union[NumberInputType, None, DatetimeRestrictedInputType]) -> Union[NumberFunctionReturnType, DatetimeFunctionReturnType]:
    """
    {
        "function": "MIN",
        "description": "Returns the minimum value among the passed arguments.",
        "search_terms": ["min", "minimum", "maximum"],
        "category": "MATH",
        "examples": [
            "MIN(10, 11)",
            "MIN(Old_Data, New_Data)"
        ],
        "syntax": "MIN(value1, [value2, ...])",
        "syntax_elements": [{
                "element": "value1",
                "description": "The first number or column to consider for the minumum value."
            },
            {
                "element": "value2, ... [OPTIONAL]",
                "description": "Additional numbers or columns to compute the minumum value from."
            }
        ]
    }
    """
    default_value = sys.maxsize
    for arg in argv:
        if isinstance(arg, pd.Series) and is_datetime_dtype(str(arg.dtype)):
            default_value = pd.Timestamp.max
            break
        elif isinstance(arg, pd.Timestamp) or isinstance(arg, datetime):
            default_value = pd.Timestamp.max
            break
        elif isinstance(arg, pd.DataFrame):
            # Check if any series are datetimes
            for col in arg.columns:
                if is_datetime_dtype(str(arg[col].dtype)):
                    default_value = pd.Timestamp.max
                    break
        elif isinstance(arg, RollingRange):
            # Check if any series are datetimes
            for col in arg.obj.columns:
                if is_datetime_dtype(str(arg.obj[col].dtype)):
                    default_value = pd.Timestamp.max
                    break

    result = get_final_result_series_or_primitive(
        default_value,
        argv,
        lambda df: df.min().min(),
        lambda previous_value, new_value: min(previous_value, new_value),
        lambda previous_series, new_series: pd.concat([previous_series, new_series], axis=1).min(axis=1)
    )

    # If we don't find any arguements, we default to 0 -- like Excel
    kept_default_min_value = not isinstance(result, pd.Series) and result == default_value
    return result if not kept_default_min_value else 0


@cast_values_in_all_args_to_type('number')
@handle_sheet_function_errors
def MULTIPLY(*argv: Optional[NumberInputType]) -> NumberFunctionReturnType:
    """
    {
        "function": "MULTIPLY",
        "description": "Returns the product of two numbers.",
        "search_terms": ["multiply", "product"],
        "category": "MATH",
        "examples": [
            "MULTIPLY(2,3)",
            "MULTIPLY(A,3)"
        ],
        "syntax": "MULTIPLY(factor1, [factor2, ...])",
        "syntax_elements": [{
                "element": "factor1",
                "description": "The first number to multiply."
            },
            {
                "element": "factor2, ... [OPTIONAL]",
                "description": "Additional numbers or series to multiply."
            }
        ]
    }
    """

    return get_final_result_series_or_primitive(
        1,
        argv,
        lambda df: df.prod().prod(),
        lambda previous_value, new_value: previous_value * new_value,
        lambda previous_series, new_series: previous_series * new_series
    )


@cast_values_in_arg_to_type('arg', 'number')
@cast_values_in_arg_to_type('power', 'number')
@handle_sheet_function_errors
def POWER(arg: NumberRestrictedInputType, power: NumberRestrictedInputType) -> NumberFunctionReturnType:
    """
    {
        "function": "POWER",
        "description": "The POWER function can be used to raise a number to a given power.",
        "search_terms": ["power", "raise", "exponent", "square", "cube"],
        "category": "MATH",
        "examples": [
            "POWER(4, 1/2)",
            "POWER(Dose, 2)"
        ],
        "syntax": "POWER(value, exponent)",
        "syntax_elements": [{
                "element": "value",
                "description": "Number to raise to a power."
            },
            {
                "element": "exponent",
                "description": "The number to raise value to."
            }
        ]
    }
    """
    if isinstance(arg, int) or isinstance(arg, float):
        return arg ** power
    
    return arg.pow(power) # type: ignore

@cast_values_in_arg_to_type('arg', 'number')
@cast_values_in_arg_to_type('decimals', 'number')
@handle_sheet_function_errors
def ROUND(arg: NumberRestrictedInputType, decimals: Optional[IntRestrictedInputType]=None) -> NumberFunctionReturnType:
    """
    {
        "function": "ROUND",
        "description": "Rounds a number to a given number of decimals.",
        "search_terms": ["round", "decimal", "integer"],
        "category": "MATH",
        "examples": [
            "ROUND(1.3)",
            "ROUND(A, 2)"
        ],
        "syntax": "ROUND(value, [decimals])",
        "syntax_elements": [{
                "element": "value",
                "description": "The value or series to round."
            },
            {
                "element": "decimals",
                "description": " The number of decimals to round to. Default is 0."
            }
        ]
    }
    """

    def excel_round(x: Optional[Union[int, float]], decimals: int) -> Union[int, float, None]:
        if x is None or pd.isna(x): 
            return None

        context = decimal.getcontext()
        context.rounding = decimal.ROUND_HALF_UP
        number = decimal.Decimal(x)
        rounded_number = round(number, decimals)
        return type(x)(rounded_number)

    # If no decimals option is passed, round to no decimals
    if decimals is None:
        decimals = 0

    if (isinstance(arg, int) or isinstance(arg, float)) and isinstance(decimals, int):
        return excel_round(arg, decimals) 

    index = get_index_from_series(arg, decimals)
    arg = get_series_from_primitive_or_series(arg, index).fillna(np.nan)
    decimals = get_series_from_primitive_or_series(decimals, index).fillna(0)

    return pd.Series(
        [excel_round(num, decimal) for num, decimal in zip(arg, decimals)], # type: ignore
        index=index
    )


@cast_values_in_arg_to_type('arg', 'number')
@handle_sheet_function_errors
def SKEW(arg: NumberRestrictedInputType) -> NumberFunctionReturnType:
    """
    {
        "function": "SKEW",
        "description": "Computes the skew of a series, excluding missing values.",
        "search_terms": [],
        "category": "MATH",
        "examples": [
            "=SKEW(A)",
            "=SKEW(A * B)"
        ],
        "syntax": "SKEW(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to calculate the skew of."
            }
        ]
    }
    """
    if isinstance(arg, int) or isinstance(arg, float):
        return 0
        
    return arg.skew() # type: ignore

@cast_values_in_arg_to_type('arg', 'number')
@handle_sheet_function_errors
def STDEV(arg: NumberInputType) -> NumberFunctionReturnType:
    """
    {
        "function": "STDEV",
        "description": "Computes the standard deviation of a series, excluding missing values.",
        "search_terms": ["standard", "deviation", "standard", "distribution"],
        "category": "MATH",
        "examples": [
            "=STDEV(A)",
            "=STDEV(A * B)"
        ],
        "syntax": "STDEV(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to calculate the standard deviation of."
            }
        ]
    }
    """
    if isinstance(arg, int) or isinstance(arg, float):
        return 0
    elif isinstance(arg, pd.Series):
        return arg.std() # type: ignore
    elif isinstance(arg, pd.DataFrame):
        return arg.stack().std() # We have to compute them all together
    else:
        return arg.apply(lambda x: x.stack().std()) # type: ignore


@cast_values_in_all_args_to_type('number')
@handle_sheet_function_errors
def SUM(*argv: Optional[NumberInputType]) -> NumberFunctionReturnType:
    """
    {
        "function": "SUM",
        "description": "Returns the sum of the given numbers and series.",
        "search_terms": ["add"],
        "category": "MATH",
        "examples": [
            "SUM(10, 11)",
            "SUM(A, B, D, F)",
            "SUM(A, B, D, F)"
        ],
        "syntax": "SUM(value1, [value2, ...])",
        "syntax_elements": [{
                "element": "value1",
                "description": "The first number or column to add together."
            },
            {
                "element": "value2, ... [OPTIONAL]",
                "description": "Additional numbers or columns to sum."
            }
        ]
    }
    """

    return get_final_result_series_or_primitive(
        0,
        argv,
        lambda df: df.sum().sum(),
        lambda previous_value, new_value: previous_value + new_value,
        lambda previous_series, new_series: previous_series + new_series
    )

@cast_values_in_all_args_to_type('number')
@handle_sheet_function_errors
def SUMPRODUCT(*argv: Union[pd.Series, pd.DataFrame]) -> NumberFunctionReturnType:
    """
    {
        "function": "SUMPRODUCT",
        "description": "Returns the sum of the product of the passed arguments.",
        "search_terms": ["sum product", "sumproduct", "sum", "product", "weighted average"],
        "category": "MATH",
        "examples": [
            "SUMPRODUCT(A:A, B:B)",
            "SUMPRODUCT(A:B)"
        ],
        "syntax": "SUMPRODUCT(array1, [array2, ...])",
        "syntax_elements": [{
                "element": "array1",
                "description": "The first array argument whose components you want to multiply and then add."
            },
            {
                "element": "value2, ... [OPTIONAL]",
                "description": "Additional series to multiply then add."
            }
        ]
    }
    """

    # We need to make sure that all of the passed arguments are the same length
    # so we can multiply them together. They must be all dataframes or all series
    for arg in argv:
        error = MitoError(
            'invalid_args_error',
            'SUMPRODUCT',
            f"SUMPRODUCT requires all arguments to be the same dimensions.",
            error_modal=False
        )

        length = len(arg)
        if length != len(argv[0]):
            raise error
        
        if type(arg) != type(argv[0]):
            raise error
        
    if isinstance(argv[0], pd.Series):
        # Put all the values into a single dataframe, prod the rows, and then
        # sum the results
        return pd.concat(argv, axis=1).fillna(1).prod(axis=1)
    else:
        # Multiply argv into a single dataframe - making the columns the same first
        result = argv[0]
        for arg in argv[1:]:
            result = result.mul(arg, fill_value=1)

        # And then sum the results
        return result.sum().sum()
        

@cast_values_in_arg_to_type('arg', 'number')
@handle_sheet_function_errors
def VALUE(arg: NumberRestrictedInputType) -> NumberFunctionReturnType:
    """
    {
        "function": "VALUE",
        "description": "Converts a string series to a number series. Any values that fail to convert will return an NaN.",
        "search_terms": ["number", "to number", "dtype", "convert", "parse"],
        "category": "MATH",
        "examples": [
            "=VALUE(A)",
            "=VALUE('123')"
        ],
        "syntax": "VALUE(string)",
        "syntax_elements": [{
                "element": "string",
                "description": "The string or series to convert to a number."
            }
        ]
    }
    """
    if isinstance(arg, int) or isinstance(arg, float):
        return arg
    
    return arg.fillna(np.nan) # type: ignore


@cast_values_in_arg_to_type('arg', 'number')
@handle_sheet_function_errors
def VAR(arg: NumberInputType) -> NumberFunctionReturnType:
    """
    {
        "function": "VAR",
        "description": "Computes the variance of a series, excluding missing values.",
        "search_terms": ["variance"],
        "category": "MATH",
        "examples": [
            "=VAR(A)",
            "=VAR(A - B)"
        ],
        "syntax": "VAR(series)",
        "syntax_elements": [{
                "element": "series",
                "description": "The series to calculate the variance of."
            }
        ]
    }
    """
    if isinstance(arg, int) or isinstance(arg, float):
        return 0
    elif isinstance(arg, pd.Series):
        return arg.var() # type: ignore
    elif isinstance(arg, pd.DataFrame):
        return arg.stack().var() # type: ignore
    else:
        return arg.apply(lambda x: x.stack().var()) # type: ignore


NUMBER_FUNCTIONS = {
    'ABS': ABS,
    'AVG': AVG,
    'CORR': CORR,
    'FLOAT': FLOAT,
    'INT': INT,
    'EXP': EXP,
    'KURT': KURT,
    'LOG': LOG,
    'MAX': MAX,
    'MIN': MIN,
    'MULTIPLY': MULTIPLY,
    'POWER': POWER,
    'ROUND': ROUND,
    'SKEW': SKEW,
    'SUM': SUM,
    'SUMPRODUCT': SUMPRODUCT,
    'STDEV': STDEV,
    'VALUE': VALUE,
    'VAR': VAR
}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import List, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State

class ResetIndexCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, sheet_index: int, drop: bool):
        super().__init__(prev_state)
        self.sheet_index = sheet_index
        self.drop = drop

        self.df_name = self.prev_state.df_names[self.sheet_index]

    def get_display_name(self) -> str:
        return 'Reset Index'
    
    def get_description_comment(self) -> str:
        return f"Reset {self.df_name} index"

    def get_code(self) -> Tuple[List[str], List[str]]:
        return [f'{self.df_name} = {self.df_name}.reset_index({"drop=True" if self.drop else ""})'], []

    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]
    


from collections import OrderedDict
import os
import pandas as pd
import pytest
from mitosheet.streamlit.v1 import spreadsheet, RunnableAnalysis
from mitosheet.tests.decorators import requires_streamlit

df1 = pd.DataFrame({'A': [123]})
df2 = pd.DataFrame({'A': [1234]})

SPREADSHEET_PARAMS = [
    (
        'Empty params',
        [], {},
        (
            {}, ''
        )
    ),
    (
        'A single dataframe',
        [df1], {},
        (
            OrderedDict({'df1': df1}), ''
        )
    ),
    (
        'Multiple dataframes',
        [df1, df2, df1], {},
        (
            OrderedDict({'df1': df1, 'df2': df2, 'df3': df1}), ''
        )
    ),
    (
        'return type of code',
        [df1], {'return_type': 'default_list'},
        (
            [df1], ''
        )
    ),
    (
        'return type of code',
        [df1], {'return_type': 'code'},
        (
            ''
        )
    ),
    (
        'return type of dfs',
        [df1], {'return_type': 'dfs_dict'},
        (
            OrderedDict({'df1': df1})
        )
    ),
    (
        'return type of dfs_list',
        [df1], {'return_type': 'dfs_list'},
        (
            [df1]
        )
    ),
    (
        'return type of analysis',
        [df1], {'return_type': 'analysis'},
        (
            RunnableAnalysis('', None, '', [])
        )
    )
]

@pytest.mark.parametrize('test, args, kwargs, expected', SPREADSHEET_PARAMS)
@requires_streamlit
def test_creates_spreadsheet(test, args, kwargs, expected):

    result = spreadsheet(*args, **kwargs)

    if isinstance(result, tuple):
        for v1, v2 in zip(result, expected):
            if isinstance(v1, OrderedDict):
                assert len(v1) == len(v2)
                for k in v1:
                    assert v1[k].equals(v2[k])
            elif isinstance(v1, list):
                for e1, e2 in zip(v1, v2):
                    assert e1.equals(e2)
            else:
                assert v1 == v2
    elif isinstance(result, list):
        for e1, e2 in zip(result, expected):
            assert e1.equals(e2)
    elif isinstance(result, OrderedDict):
        assert len(result) == len(expected)
        for k in result:
            assert result[k].equals(expected[k])
    elif isinstance(result, RunnableAnalysis):
        assert isinstance(expected, RunnableAnalysis)
    else:
        assert result == expected

@requires_streamlit
def test_return_type_function():
    f = spreadsheet(df1, code_options={'as_function': True, 'call_function': False, 'function_name': 'test', 'function_params': {}}, return_type='function')
    assert callable(f)

@requires_streamlit
def test_return_type_function_invalid_code_options():
    with pytest.raises(ValueError):
        spreadsheet(df1, return_type='function')
    with pytest.raises(ValueError):
        spreadsheet(df1, code_options={'as_function': False, 'call_function': False, 'function_name': 'test', 'function_params': {}}, return_type='function')
    with pytest.raises(ValueError):
        spreadsheet(df1, code_options={'as_function': True, 'call_function': True, 'function_name': 'test', 'function_params': {}}, return_type='function')



from dash import Dash, callback, Input, Output, html, dcc
from mitosheet.mito_dash.v1 import Spreadsheet, mito_callback, activate_mito
import pandas as pd

app = Dash(__name__)
activate_mito(app)

app.layout = html.Div([
    html.H1("Stock Analysis", style={'color': 'white'}),
    # A dropdown for selecting between two dataframes
    dcc.Dropdown(
        id='dropdown',
        options=[
            {'label': 'Stock Data', 'value': 'stock_data'},
            {'label': 'Stock Data 2', 'value': 'stock_data_2'},
        ],
        value='stock_data'
    ),
    Spreadsheet(id={'type': 'spreadsheet', 'id': 'sheet'}),
])

@callback(
    Output({'type': 'spreadsheet', 'id': 'sheet'}, 'data'),
    Input('dropdown', 'value')
)
def update_spreadsheet_data(dropdown_value):
    if dropdown_value == 'stock_data':
        df = pd.DataFrame({'First Data': [1, 2, 3], 'Second Data': [4, 5, 6]})
    else:
        df = pd.DataFrame({'First Data': [7, 8, 9], 'Second Data': [10, 11, 12]})

    return df.to_dict('records')

if __name__ == '__main__':
    app.run_server(debug=True)


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import io
from typing import Any, Dict, List, Optional

import plotly.graph_objects as go

from mitosheet.state import State
from mitosheet.types import ColumnHeader

# Graph types should be kept consistent with the GraphType in GraphSidebar.tsx
SCATTER = "scatter"
LINE = "line"
BAR = "bar"
HISTOGRAM = "histogram"
BOX = "box"
STRIP = "strip"
VIOLIN = "violin"
ECDF = "ecdf"
DENSITY_HEATMAP = "density heatmap"
DENSITY_CONTOUR = "density contour"

# Label for each type of graph used in the graph title
GRAPH_TITLE_LABELS = {
    SCATTER: "scatter plot",
    LINE: "line",
    BAR: "bar chart",
    BOX: "box plot",
    HISTOGRAM: "histogram",
    STRIP: "strip",
    VIOLIN: "violin",
    ECDF: "ecdf",
    DENSITY_HEATMAP: "density heatmap",
    DENSITY_CONTOUR: "density contour",
}


def get_graph_title(
    x_axis_column_headers: List[ColumnHeader],
    y_axis_column_headers: List[ColumnHeader],
    filtered: bool,
    graph_type: str,
) -> str:
    """
    Helper function for determing the title of the graph
    """
    # Get the label to let the user know that their graph had a filter applied.
    graph_filter_label: Optional[str] = "(first 1000 rows)" if filtered else None

    # Compile all of the column headers into one comma separated string
    all_column_headers = (", ").join(
        str(s) for s in x_axis_column_headers + y_axis_column_headers
    )
    # Get the title of the graph based on the type of graph
    graph_title_label = GRAPH_TITLE_LABELS[graph_type]

    # Combine all of the non empty graph title components into one list
    graph_title_components = (
        [all_column_headers, graph_filter_label, graph_title_label]
        if graph_filter_label is not None
        else [all_column_headers, graph_title_label]
    )

    # Return a string with all of the graph_title_components separated by a space
    return " ".join(graph_title_components)


def get_new_graph_tab_name(graph_data_dict: Dict[str, Dict[str, Any]]) -> str:
    """
    Creates the name for the new graph tab sheet using the format
    graph0, graph1, etc.
    """
    all_graph_names = []
    for graph_data in graph_data_dict.values():
        all_graph_names.append(graph_data["graphTabName"])
    
    graph_number_indicator = 0
    new_graph_name = f'graph{graph_number_indicator}'

    while new_graph_name in all_graph_names:
        graph_number_indicator += 1
        new_graph_name = f'graph{graph_number_indicator}'
        
    return new_graph_name


def get_html_and_script_from_figure(
    fig: go.Figure, height: int, width: int,
    include_plotlyjs: bool,
) -> Dict[str, str]:
    """
    Given a plotly figure, generates HTML from it, and returns
    a dictonary with the div and script for the frontend.

    The plotly HTML generated by the write_html function call is a div with two children:
    1. a div that contains the id for the graph itself
    2. a script that actually builds the graph

    Because we have to dynamically execute the script, we split these into two
    strings, to make them easier to do what we need on the frontend
    """
    # Send the graph back to the frontend
    buffer = io.StringIO()
    fig.write_html(
        buffer,
        full_html=False,
        include_plotlyjs=include_plotlyjs,
        default_height=height,
        default_width=width,
    )

    original_html = buffer.getvalue()

    # Everything in a script tag we want to treat as one big script

    # The get the graph div, which looks like: <div id="9c21d143-3fcd-4958-9295-ac7ada668186" class="plotly-graph-div" style="height:425px; width:970px;"></div>
    # This is the only div we need to put on the frontend.
    div = original_html[original_html.find('<div id='):original_html.find('</div>', original_html.find('<div id='))]

    # Get all the scripts that are between the <script> tags, and join them together. If include_plotlyjs is true, this includes
    # the plotly js script, and otherwise, it's just the data and the script that uses Plotly to build the graph on the div above
    script = ''
    index = 0
    while index < len(original_html):
        script_start = original_html.find('<script type="text/javascript">', index)
        script_end = original_html.find('</script>', script_start)

        if script_start == -1 or script_end == -1:
            break

        # Get rid of the initial script
        script_start += len('<script type="text/javascript">')
        script = script + original_html[script_start:script_end] + ';\n'
        index = script_end + 9

    return {"html": div, "script": script}

def get_column_header_from_optional_column_id_graph_param(
    state: State, 
    graph_creation_params: Dict[str, Any], 
    param_name: str
) -> Optional[ColumnHeader]:
    sheet_index = graph_creation_params['sheet_index']
    if param_name in graph_creation_params.keys() and graph_creation_params[param_name] in state.column_ids.column_id_to_column_header[sheet_index].keys():
        return state.column_ids.get_column_header_by_id(sheet_index, graph_creation_params[param_name])
    else: return None




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
File that holds all the type conversion functions
"""
from typing import Any, Callable

import pandas as pd
from mitosheet.public.v1.sheet_functions.types.to_boolean_series import to_boolean_series
from mitosheet.public.v1.sheet_functions.types.to_datetime_series import \
    to_datetime_series
from mitosheet.public.v1.sheet_functions.types.to_float_series import to_float_series
from mitosheet.public.v1.sheet_functions.types.to_int_series import to_int_series
from mitosheet.public.v1.sheet_functions.types.to_string_series import to_string_series
from mitosheet.public.v1.sheet_functions.types.to_timedelta_series import \
    to_timedelta_series
from mitosheet.is_type_utils import (is_bool_dtype,
                                                   is_datetime_dtype,
                                                   is_float_dtype,
                                                   is_int_dtype,
                                                   is_timedelta_dtype)
from mitosheet.public.v1.sheet_functions.types.utils import get_datetime_format

def to_series(ele: Any) -> pd.Series:
    return pd.Series([ele])

def get_function_to_convert_to_series(column_dtype: str) -> Callable:
    if is_bool_dtype(column_dtype):
        return to_boolean_series
    elif is_datetime_dtype(column_dtype):
        return to_datetime_series
    elif is_int_dtype(column_dtype):
        return to_int_series
    elif is_float_dtype(column_dtype):
        return to_float_series
    elif is_timedelta_dtype(column_dtype):
        return to_timedelta_series
    elif column_dtype == 'series':
        return to_series 
    
    # DEfault to string series
    return to_string_series




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List, Optional, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State
from mitosheet.types import UserDefinedFunctionParamType
from mitosheet.step_performers.utils.user_defined_function_utils import get_transpiled_user_defined_function_params


class UserDefinedImportCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, importer: str, user_defined_function_params: Dict[str, Tuple[UserDefinedFunctionParamType, Any, Any]], new_df_names: List[str]):
        super().__init__(prev_state)
        self.importer = importer
        self.user_defined_function_params = user_defined_function_params

        self.df_names = new_df_names

    def get_display_name(self) -> str:
        return 'User Defined Import'
    
    def get_description_comment(self) -> str:
        return f"Imported {', '.join(self.df_names)} using {self.importer}"

    def get_code(self) -> Tuple[List[str], List[str]]:
        # For each new dataframe, we get it's name
        df_name_string = ', '.join(self.df_names)
        code = f"{df_name_string} = {self.importer}({get_transpiled_user_defined_function_params(self.user_defined_function_params)})"
        return [code], []
    
    def get_created_sheet_indexes(self) -> Optional[List[int]]:
        return [i for i in range(len(self.prev_state.dfs), len(self.prev_state.dfs) + len(self.df_names))]


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict, List, Optional, Set, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.no_op_code_chunk import NoOpCodeChunk
from mitosheet.code_chunks.step_performers.column_steps.rename_columns_code_chunk import \
    RenameColumnsCodeChunk
from mitosheet.errors import make_column_exists_error
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param


class RenameColumnStepPerformer(StepPerformer):
    """"
    A rename_column step, which allows you to rename a column
    in a dataframe.

    NOTE: this should only be called on dataframes that do
    not have multi-index headers!
    """

    @classmethod
    def step_version(cls) -> int:
        return 2

    @classmethod
    def step_type(cls) -> str:
        return 'rename_column' 

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:

        new_column_header = params['new_column_header']

        if new_column_header in prev_state.dfs[params['sheet_index']].columns:
            raise make_column_exists_error(new_column_header)

        execution_data = {
            'column_ids_to_new_column_headers': {
                params['column_id']: new_column_header
            }
        }

        return cls.execute_through_transpile(
            prev_state, 
            params, 
            execution_data
        )

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        if params['new_column_header'] == '':
            # If the new column header is an empty string, it's a noop
            return [NoOpCodeChunk(prev_state)]


        return [
            RenameColumnsCodeChunk(
                prev_state, 
                # We construct a rename for mulitple columns, as this is the most 
                # convenient way to allow us to combine multiple renames
                # into one
                params['sheet_index'],
                get_param(execution_data if execution_data is not None else {}, 'column_ids_to_new_column_headers'),
            )
        ]
    
    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains the exports for update events, which are events that manipulate
the steps in some way, but do not _just generate a step_ (see ../steps for
more information about steps).

Examples of this sort of event are:
- Lots of steps being replayed from the front-end.
- A user undoing an existing step. 
"""

from mitosheet.updates.code_options_update import CODE_OPTIONS_UPDATE
from mitosheet.updates.undo import UNDO_UPDATE
from mitosheet.updates.redo import REDO_UPDATE
from mitosheet.updates.clear import CLEAR_UPDATE
from mitosheet.updates.args_update import ARGS_UPDATE
from mitosheet.updates.save_analysis import SAVE_ANALYSIS_UPDATE
from mitosheet.updates.replay_analysis import REPLAY_ANALYSIS_UPDATE
from mitosheet.updates.set_user_field_update import SET_USER_FIELD_UPDATE
from mitosheet.updates.checkout_step_by_idx import CHECKOUT_STEP_BY_IDX_UPDATE
from mitosheet.updates.append_user_field import APPEND_USER_FIELD_UPDATE
from mitosheet.updates.update_feedback_v2_object import UPDATE_FEEDBACK_V2_OBJECT_UPDATE
from mitosheet.updates.go_pro import GO_PRO_UPDATE
from mitosheet.updates.render_count import RENDER_COUNT_UPDATE
from mitosheet.updates.checklist_update import CHECKLIST_UPDATE
from mitosheet.updates.update_existing_imports import UPDATE_EXISTING_IMPORTS_UPDATE
from mitosheet.pro.updates.undo_to_step_index import UNDO_TO_STEP_INDEX_UPDATE


# All update events must be listed in this variable.
UPDATES = [
    UNDO_UPDATE,
    REDO_UPDATE,
    CLEAR_UPDATE,
    ARGS_UPDATE,
    SAVE_ANALYSIS_UPDATE,
    REPLAY_ANALYSIS_UPDATE,
    CHECKOUT_STEP_BY_IDX_UPDATE,
    APPEND_USER_FIELD_UPDATE,
    SET_USER_FIELD_UPDATE,
    CHECKOUT_STEP_BY_IDX_UPDATE,
    UPDATE_FEEDBACK_V2_OBJECT_UPDATE,
    GO_PRO_UPDATE,
    RENDER_COUNT_UPDATE,
    CHECKLIST_UPDATE,
    UPDATE_EXISTING_IMPORTS_UPDATE,
    CODE_OPTIONS_UPDATE,
    UNDO_TO_STEP_INDEX_UPDATE
]


import inspect
from functools import wraps
from typing import Callable, List

from mitosheet.public.v3.types.utils import get_arg_cast_to_type
from mitosheet.types import PrimitiveTypeName


def cast_values_in_all_args_to_type(
    target_primitive_type_name: PrimitiveTypeName,
    primitive_types_to_ignore: List[PrimitiveTypeName] = []
) -> Callable:

    def wrap(sheet_function):
        @wraps(sheet_function)
        def wrapped_sheet_function(*args):   

            # For every arguement, go through and cast them to the correct type
            final_args = [
                get_arg_cast_to_type(target_primitive_type_name, arg, primitive_types_to_ignore=primitive_types_to_ignore)
                for arg in args
            ]

            return sheet_function(*final_args)        
        return wrapped_sheet_function
    return wrap


def cast_values_in_arg_to_type(
    arg_name: str,
    target_primitive_type_name: PrimitiveTypeName,
) -> Callable:
    def wrap(sheet_function):

        @wraps(sheet_function)
        def wrapped_sheet_function(*args):   

            arg_names = list(inspect.signature(sheet_function).parameters.keys())
            arg_index = arg_names.index(arg_name)

            final_args = [
                get_arg_cast_to_type(target_primitive_type_name, arg) if index == arg_index else arg
                for index, arg in enumerate(args)
            ]

            return sheet_function(*final_args)        
        return wrapped_sheet_function
    return wrap


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from collections import OrderedDict
from copy import deepcopy
from typing import Any, Callable, Collection, List, Dict, Optional, Set, Union
import pandas as pd

from mitosheet.column_headers import ColumnIDMap
from mitosheet.types import FrontendFormulaAndLocation, OverwriteSheetIndexParams
from mitosheet.types import ColumnHeader, ColumnID, DataframeFormat
from mitosheet.utils import  get_first_unused_dataframe_name

# Constants for where the dataframe in the state came from
DATAFRAME_SOURCE_PASSED = "passed"  # passed in mitosheet.sheet
DATAFRAME_SOURCE_IMPORTED = "imported"  # imported through a simple import
DATAFRAME_SOURCE_PIVOTED = "pivoted"  # created through a pivot
DATAFRAME_SOURCE_MERGED = "merged"  # created through a merge
DATAFRAME_SOURCE_CONCAT = "concated"  # created through a concat
DATAFRAME_SOURCE_DUPLICATED = "duplicated"  # created through a sheet duplication
DATAFRAME_SOURCE_TRANSPOSED = "transposed"  # created through a dataframe being transposed
DATAFRAME_SOURCE_MELTED = "melted"  # created through a dataframe being melted
DATAFRAME_SOURCE_AI = "ai"  # through AI generated code

# Constants used for formatting. Defined here to avoid circular imports
NUMBER_FORMAT_PLAIN_TEXT = "plain text"
NUMBER_FORMAT_CURRENCY = "currency"
NUMBER_FORMAT_ACCOUNTING = "accounting"
NUMBER_FORMAT_PERCENTAGE = "percentage"
NUMBER_FORMAT_SCIENTIFIC_NOTATION = "scientific notation"


def get_default_dataframe_format() -> DataframeFormat:
    return {
        "columns": {},
        "headers": {},
        "rows": {
            "even": {},
            "odd": {},
        },
        "border": {},
        "conditional_formats": []
    }


class State:
    """
    State is a container that stores the current state of a Mito analysis,
    where each step that is applied takes a state as input and creates a
    new state as output.

    It stores the obvious things, like the dataframes and their names, but
    also other helper pieces of state like: the column formulas, the filters,
    etc.
    """

    def __init__(
        self,
        dfs: Collection[pd.DataFrame],
        public_interface_version: int,
        df_names: Optional[List[str]]=None,
        df_sources: Optional[List[str]]=None,
        column_ids: Optional[ColumnIDMap]=None,
        column_formulas: Optional[List[Dict[ColumnID, List[FrontendFormulaAndLocation]]]]=None,
        column_filters: Optional[List[Dict[ColumnID, Any]]]=None,
        df_formats: Optional[List[DataframeFormat]]=None,
        graph_data_dict: "Optional[OrderedDict[str, Dict[str, Any]]]"=None,
        user_defined_functions: Optional[List[Callable]]=None,
        user_defined_importers: Optional[List[Callable]]=None,
        user_defined_editors: Optional[List[Callable]]=None,
    ):

        # The dataframes that are in the state
        self.dfs = list(dfs)

        self.public_interface_version = public_interface_version

        # The df_names are composed of two parts:
        # 1. The names of the variables passed into the mitosheet.sheet call (which don't change over time).
        # 2. The names of the dataframes that were created during the analysis (e.g. by a merge).
        # Until we get them from the frontend as an update_event, we default them to df1, df2, ...
        self.df_names = (
            df_names
            if df_names is not None
            else [f"df{i + 1}" for i in range(len(dfs))]
        )

        # The df sources are where the actual dataframes come from, e.g.
        # how the dataframe was created. If not df sources passed, then this is in the
        # initialize state, and so these dataframes were passed to the mitosheet
        # call
        self.df_sources = (
            df_sources
            if df_sources is not None
            else [DATAFRAME_SOURCE_PASSED for _ in dfs]
        )

        # We then make a column id map if we do not already have one, so that we can identify each
        # of the columns from their static ids through the rest of the analysis.
        # NOTE: every state variable below that is defined per column is access by _ids_ not
        # by column headers. The column headers _only_ index into the dataframe itself
        self.column_ids = column_ids if column_ids else ColumnIDMap(dfs)

        self.column_formulas: List[Dict[ColumnID, List[FrontendFormulaAndLocation]]] = (
            column_formulas
            if column_formulas is not None
            else [
                {
                    column_id: []
                    for column_id in self.column_ids.get_column_ids(sheet_index)
                }
                for sheet_index in range(len(dfs))
            ]
        )

        self.column_filters = (
            column_filters
            if column_filters is not None
            else [
                {
                    column_id: {"operator": "And", "filters": []}
                    for column_id in self.column_ids.get_column_ids(sheet_index)
                }
                for sheet_index in range(len(dfs))
            ]
        )

        self.df_formats: List[DataframeFormat] = (
            df_formats
            if df_formats is not None
            else [
                get_default_dataframe_format()
                for _ in range(len(dfs))
            ]
        )

        # We put this in an ordered dict so we can easily figure out the last graph that was edited at each step. 
        # This is helpful for undoing, for example. 
        self.graph_data_dict: OrderedDict[str, Dict[str, Any]] = graph_data_dict if graph_data_dict is not None else OrderedDict()

        self.user_defined_functions = user_defined_functions if user_defined_functions is not None else []
        self.user_defined_importers = user_defined_importers if user_defined_importers is not None else []
        self.user_defined_editors = user_defined_editors if user_defined_editors is not None else []

    def copy(self, deep_sheet_indexes: Optional[Union[List[int], Set[int], None]]=None) -> "State":
        """
        Returns a copy of the state, while only making deep copies of
        those dataframes in the deep_sheet_indexes. Ideally, we'd copy
        even less than that deeply.
        """
        if deep_sheet_indexes is None:
            deep_sheet_indexes = []
        
        return State(
            [df.copy(deep=index in deep_sheet_indexes) for index, df in enumerate(self.dfs)],
            self.public_interface_version,
            df_names=deepcopy(self.df_names),
            df_sources=deepcopy(self.df_sources),
            column_ids=deepcopy(self.column_ids),
            column_formulas=deepcopy(self.column_formulas),
            column_filters=deepcopy(self.column_filters),
            df_formats=deepcopy(self.df_formats),
            graph_data_dict=deepcopy(self.graph_data_dict),
            user_defined_functions=deepcopy(self.user_defined_functions),
            user_defined_importers=deepcopy(self.user_defined_importers),
            user_defined_editors=deepcopy(self.user_defined_editors),
        )

    def add_df_to_state(
        self,
        new_df: pd.DataFrame,
        df_source: str,
        df_name: Optional[str]=None,
        df_format: Optional[DataframeFormat]=None,
        use_deprecated_id_algorithm: bool=False, 
        overwrite: Optional[OverwriteSheetIndexParams]=None,
    ) -> int:
        """
        Helper function for adding a new dataframe to this state,
        and keeping all the other variables in sync.

        If sheet_index is defined, then will replace the dataframe
        that is currently at the index. Otherwise, if sheet_index is
        not defined, then will append the df to the end of the state
        """
        if overwrite is None:
            # Update dfs by appending new df
            self.dfs.append(new_df)
            # Also update the dataframe name
            if df_name is None:
                self.df_names.append(
                    get_first_unused_dataframe_name(self.df_names, f"df{len(self.df_names) + 1}")
                )
            else:
                self.df_names.append(df_name)

            # Save the source of this dataframe
            self.df_sources.append(df_source)

            # Add this to the column_ids map
            column_ids = self.column_ids.add_df(
                new_df, use_deprecated_id_algorithm=use_deprecated_id_algorithm
            )

            # Update all the variables that depend on column_headers
            self.column_formulas.append(
                {column_id: [] for column_id in column_ids}
            )

            self.column_filters.append(
                {
                    column_id: {"operator": "And", "filters": []}
                    for column_id in column_ids
                }
            )
            self.df_formats.append(
                get_default_dataframe_format()
                if df_format is None
                else df_format
            )

            # Return the index of this sheet
            return len(self.dfs) - 1
        else:
            sheet_index = overwrite['sheet_index_to_overwrite']
            attempt_to_save_filter_metadata = overwrite['attempt_to_save_filter_metadata']

            # Update dfs by switching which df is at this index specifically
            self.dfs[sheet_index] = new_df
            # Also update the dataframe name, if it is passed. Otherwise, we don't change it
            if df_name is not None:
                self.df_names[sheet_index] = df_name

            # Save the source of this dataframe, if it is passed. Otherwise, don't change it
            if df_source is not None:
                self.df_sources[sheet_index] = df_source

            # Add this to the column_ids map
            column_ids = self.column_ids.add_df(
                new_df,
                sheet_index=sheet_index,
                use_deprecated_id_algorithm=use_deprecated_id_algorithm,
            )

            # Update all the variables that depend on column_headers
            self.column_formulas[sheet_index] = {
                column_id: [] for column_id in column_ids
            }

            if not attempt_to_save_filter_metadata:
                self.column_filters[sheet_index] = {
                    column_id: {"operator": "And", "filters": []}
                    for column_id in column_ids
                }
            else:
                # Save the filter metadata for any column id that still exists
                old_column_filters = self.column_filters[sheet_index]
                self.column_filters[sheet_index] = {
                    column_id: (
                        {"operator": "And", "filters": []} if column_id not in old_column_filters else
                        old_column_filters[column_id])
                    for column_id in column_ids
                    
                }

            self.df_formats[sheet_index] = (
                get_default_dataframe_format()
                if df_format is None
                else df_format
            )

            # Return the index of this sheet
            return sheet_index

    def add_columns_to_state(self, sheet_index: int, column_headers: List[ColumnHeader], column_headers_to_column_ids: Optional[Dict[ColumnHeader, ColumnID]]=None) -> List[ColumnID]:
        """
        Helper function for adding a new columns to this state, making sure that we 
        track the relevant metadata variables.
        """
        # Update column state variables
        new_column_ids = []
        for column_header in column_headers:
            column_id = column_headers_to_column_ids.get(column_header, None) if column_headers_to_column_ids else None
            column_id = self.column_ids.add_column_header(sheet_index, column_header, column_id=column_id)
            self.column_formulas[sheet_index][column_id] = []
            self.column_filters[sheet_index][column_id] = {'operator': 'And', 'filters': []}
            new_column_ids.append(column_id)
        return new_column_ids


    def does_sheet_index_exist_within_state(self, sheet_index: int) -> bool:
        """
        Returns true iff a sheet_index exists within this state
        """
        return not (sheet_index < 0 or sheet_index >= len(self.dfs))

    def move_to_deprecated_id_algorithm(self) -> None:
        """
        This helper function will move the entire state to the new IDs,
        for backwards compatibility reasons. Namely, users who pass
        dataframes to mito directly have a bulk_old_rename appended to
        the start of their analysis that calls this function.

        This moves the IDs to what they would have been if there was a
        preprocessing step that had run, that had performed renames on
        the column headers before they were allowed into Mito. Brutal.
        """
        from mitosheet.step_performers.bulk_old_rename.deprecated_utils import (
            make_valid_header,
        )

        # Loop over all the attributes of this object
        for key, value in self.__dict__.items():
            # And for anything defined on columns, update it to the new id schema
            if key.startswith("column") and key != "column_ids":
                new_value = [
                    {make_valid_header(k): v for k, v in column_map.items()}
                    for column_map in value
                ]
                self.__setattr__(key, new_value)

        # Then, update the column ids mapping object itself
        self.column_ids.move_to_deprecated_id_format()


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict, List, Optional, Set, Type
import json
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.column_steps.set_column_formula import SetColumnFormulaStepPerformer
from mitosheet.step_performers.filter import FilterStepPerformer
from mitosheet.state import State
from mitosheet.step_performers import STEP_TYPE_TO_STEP_PERFORMER
from mitosheet.types import FORMULA_SPECIFIC_INDEX_LABELS_TYPE, ColumnHeader, ColumnID, FORMULA_ENTIRE_COLUMN_TYPE


class Step:
    """
    A step is a container around a specific data transformation.

    A step has parameters - e.g. which sheet it touches, which column
    it edits.

    It then is linked to a StepPerformer class via the step_type. 
    The Step uses a mapping from its step_type to the StepPerformer 
    to figure out the correct step generator to use at any given time. 

    This StepPerformer is responsible for "executing" this step, which 
    functionally means taking the prev_state that this step starts at,
    and using the step's parameters to create a new post_state that 
    is up to date.

    This also stores execution_data from that execution - which makes
    for easy transpiling of the step! For example, a simple import
    stores what the delimeters used in the files it read in are, so
    that it can transpile this easily.
    """

    def __init__(
        self, 
        step_type: str, 
        step_id: str, 
        params: Dict[str, Any],
        prev_state: Optional[State] = None,
        post_state: Optional[State] = None,
        execution_data: Optional[Dict[str, Any]] = None,
    ):
        # The type of the step, one for each type of transformation
        self.step_type = step_type
        # The id of the step, so that steps can overwrite eachother by id
        self.step_id = step_id
        # The parameters that the step takes as input - which in turn are used
        # in execution, transpilation, and in describing the step
        # NOTE: these must be saturated before they can be used!
        self.params = params

        # The state at the start of this step; is None only for the initialize step
        self.prev_state = prev_state
        # The state you get from executing the prev_state with the passed params
        self.post_state = post_state
        # execution_data is data from the execution of the data transformation that
        # is useful for the transpiler - that means the transpiler can do way less
        # work if it has already been done. See simple_import for an example
        self.execution_data = execution_data if execution_data is not None else {}


    @property
    def dfs(self):
        return self.post_state.dfs

    @property
    def df_names(self):
        return self.post_state.df_names

    @property
    def df_sources(self):
        return self.post_state.df_sources

    @property
    def column_ids(self):
        return self.post_state.column_ids
    
    @property
    def column_formulas(self):
        return self.post_state.column_formulas
    
    @property
    def column_filters(self):
        return self.post_state.column_filters

    @property
    def df_formats(self):
        return self.post_state.df_formats

    @property
    def graph_data_dict(self) -> Dict[str, Dict[str, Any]]:
        """
        graph_data_dict contains all of the parameters used to construct the graph,
        the actual graph html & javascript, and the generated code for all of the existing graphs in Mito.
        """
        return self.final_defined_state.graph_data_dict

    @property
    def step_performer(self) -> Type[StepPerformer]:
        """
        A helper function to return the step performer class
        that is associated with this step, based on step type.

        This is useful for other pieces of the codebase
        that need the step performer.
        """
        step_performer = STEP_TYPE_TO_STEP_PERFORMER[self.step_type]
        return step_performer

    @property
    def initial_defined_state(self) -> State:
        """
        Returns the initial defined state in this step, as the prev and post
        state are optional, but we also need a step to have a defined state
        """
        return self.prev_state if self.prev_state is not None else \
            (self.post_state if self.post_state is not None else State([], public_interface_version=1))

    @property
    def final_defined_state(self) -> State:
        """
        Returns the final defined state in this step, as the prev and post
        state are optional, but we also need a step to have a defined state
        """
        return self.post_state if self.post_state is not None else \
            (self.prev_state if self.prev_state is not None else State([], 1))

    def set_prev_state_and_execute(self, new_prev_state: State, previous_steps: List["Step"]) -> bool:
        """
        Changes the prev_state of this step, which in turns triggers
        a reexecution with the same parameters. 

        If successful, will update the step in-place. If it fails, 
        this will not update the step.

        NOTE: this is the only function you should use to get a step
        to execute!

        Returns True if the step returns a new post_state, meaning an
        execution actually occured.
        """        
        # Saturate the event to get up to date parameters
        # TODO: this should fill in the execution data - hopefully
        # we can get all of it without executing. I think we probably can
        params = self.step_performer.saturate(new_prev_state, self.params, previous_steps)

        # Actually execute the data transformation
        post_state_and_execution_data = self.step_performer.execute(new_prev_state, params)

        if post_state_and_execution_data is not None:
            # If we don't get anything new back, then we just make this
            # step a no-op, and don't do anything
            (new_post_state, execution_data) = post_state_and_execution_data

        else:
            # Sometimes step execution returns None, which functionally means that 
            # nothing changed in the step (but we need not error). In this case, we
            # just don't change anything in the state
            new_post_state, execution_data = new_prev_state, {}
        
        # Update the relevant state variables
        self.prev_state = new_prev_state
        self.post_state = new_post_state
        self.execution_data = execution_data if execution_data is not None else {}
        self.params = params

        return post_state_and_execution_data is not None
    

    def step_indexes_to_skip(self, all_steps_before_this_step: List['Step']) -> Set[int]:
        """
        Given the steps that come before it, any step has the ability
        to skip these steps. 

        Currently, a step only skips the steps before it if:
        1. This step is a filter step that is trying to replace an older filter step
        2. This step has the same id as any step before it (like for pivot tables)
        3. This step is a formula step overwriting the step that came just before it AND they both set the entire column
        4. This step is a formula step overwriting the step that came just before it AND they both set the same indexes
        """

        step_indexes_to_skip = set()

        for step_index, step in enumerate(all_steps_before_this_step):
            # Check (1)
            if step.step_type == FilterStepPerformer.step_type() and self.step_type == FilterStepPerformer.step_type():
                if step.params['sheet_index'] == self.params['sheet_index'] \
                    and step.params['column_id'] == self.params['column_id']:
                   step_indexes_to_skip.add(step_index) 
            
            # Check (2)
            elif step.step_id == self.step_id:
                step_indexes_to_skip.add(step_index)

        if len(all_steps_before_this_step) > 0:
            previous_step: Step = all_steps_before_this_step[-1]
            
            # Check (3) and (4)
            if self.step_type == SetColumnFormulaStepPerformer.step_type() and previous_step.step_type == SetColumnFormulaStepPerformer.step_type():
                both_entire_column = self.params['index_labels_formula_is_applied_to']['type'] == FORMULA_ENTIRE_COLUMN_TYPE and previous_step.params['index_labels_formula_is_applied_to']['type'] == FORMULA_ENTIRE_COLUMN_TYPE
                same_indexes = (
                    self.params['index_labels_formula_is_applied_to']['type'] == FORMULA_SPECIFIC_INDEX_LABELS_TYPE and previous_step.params['index_labels_formula_is_applied_to']['type'] == FORMULA_SPECIFIC_INDEX_LABELS_TYPE \
                    and self.params['index_labels_formula_is_applied_to']['index_labels'] == previous_step.params['index_labels_formula_is_applied_to']['index_labels']
                )
                
                if (both_entire_column or same_indexes) \
                    and self.params['sheet_index'] == previous_step.params['sheet_index'] \
                    and self.params['column_id'] == previous_step.params['column_id']:
                    step_indexes_to_skip.add(len(all_steps_before_this_step) - 1)

        return step_indexes_to_skip

    def get_column_headers_by_ids(self, sheet_index: int, column_ids: List[ColumnID]) -> List[Any]:
        """
        Utility for getting the column headers from column ids in a step.
        First attempts to get them from the prev state, but if not 
        """
        return self.final_defined_state.column_ids.get_column_headers_by_ids(sheet_index, column_ids)
            
    def get_column_header_by_id(self, sheet_index: int, column_id: ColumnID) -> Any:
        """
        Utility for getting the column header from a column id in a step.
        First attempts to get them from the prev state, but if not 
        """
        return self.final_defined_state.column_ids.get_column_header_by_id(sheet_index, column_id)

    def get_column_id_by_header(self, sheet_index: int, column_header: ColumnHeader) -> str:
        """
        Utility for getting the column id from a column header in a step.
        First attempts to get them from the prev state, but if not 
        """
        return self.final_defined_state.column_ids.get_column_id_by_header(sheet_index, column_header)

    def __repr__(self) -> str:
        return json.dumps({
            'step_type': self.step_type,
            'params': self.params
        })



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Mito.
# Distributed under the terms of the Modified BSD License.
import collections
from typing import Any, Dict, Collection, List, Optional, Tuple
import pandas as pd
from mitosheet.errors import make_duplicated_column_headers_error
from mitosheet.telemetry.telemetry_utils import log
from mitosheet.preprocessing.preprocess_step_performer import \
    PreprocessStepPerformer
from mitosheet.types import StepsManagerType


class CheckArgsTypePreprocessStepPerformer(PreprocessStepPerformer):
    """
    This preprocessing step is responsible for making sure the arguments
    passed are of the right type.
    """

    @classmethod
    def preprocess_step_version(cls) -> int:
        return 1

    @classmethod
    def preprocess_step_type(cls) -> str:
        return 'check_args_type'

    @classmethod
    def execute(cls, args: Collection[Any]) -> Tuple[List[Any], Optional[List[str]], Optional[Dict[str, Any]]]:
        # We first validate all the parameters as either dataframes or strings
        # but we also allow users to pass None values, which we just ignore (this
        # makes variable number of inputs to the sheet possible).
        
        # NOTE: if passing None values to have them ignored, the user should 
        # be careful to place them at the end of the arguments (so df names)
        # read properly and match up with the correct variables
        for arg in args:
            if not isinstance(arg, pd.DataFrame) \
                and not isinstance(arg, pd.Series) \
                and not isinstance(arg, str) \
                and not isinstance(arg, dict) \
                and not isinstance(arg, list) \
                and not arg is None:
                error_message = f'Invalid argument passed to sheet: {arg}. Please pass all dataframes or paths to CSV files.'
                log('mitosheet_sheet_call_failed', {'error': error_message}, failed=True)
                raise ValueError(error_message)

            # We also validate there are no duplicated headers in dataframes
            if isinstance(arg, pd.DataFrame):
                duplicated_headers = [item for item, count in collections.Counter(arg.columns).items() if count > 1]
                if len(duplicated_headers) > 0:
                    raise make_duplicated_column_headers_error(duplicated_headers)
        
        # We do filter out all the None arguments
        return [arg for arg in args if arg is not None], None, None

    @classmethod
    def transpile(cls, steps_manager: StepsManagerType, execution_data: Optional[Dict[str, Any]]) -> Tuple[List[str], List[str]]:
        return [], []



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List, Optional, Set

import pandas as pd

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.column_steps.reorder_column_code_chunk import \
    ReorderColumnCodeChunk
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param


def get_valid_index(dfs: List[pd.DataFrame], sheet_index: int, new_column_index: int) -> int:
    # make sure new_column_index is valid
    if new_column_index < 0:
        new_column_index = 0

    if new_column_index >= len(dfs[sheet_index].columns):
        new_column_index = len(dfs[sheet_index].columns) - 1

    return new_column_index


class ReorderColumnStepPerformer(StepPerformer):
    """""
    A reorder_column step, which allows you to move 
    a column to a different location in the df.
    """

    @classmethod
    def step_version(cls) -> int:
        return 2

    @classmethod
    def step_type(cls) -> str:
        return 'reorder_column'

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            ReorderColumnCodeChunk(
                prev_state, 
                get_param(params, 'sheet_index'),
                get_param(params, 'column_id'),
                get_param(params, 'new_column_index'),
            )
        ]
    
    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List
from mitosheet.saved_analyses.step_upgraders.utils_column_header_to_column_id import \
    replace_headers_with_id


def upgrade_set_column_formula_1_to_2(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Moves to using column id instead of column header.

    OLD: {
        'step_version': 1, 
        'step_type': "set_column_formula",
        'params': {
            sheet_index: 0,
            column_header: _header_,
            old_formula: '=A',
            new_formula: '=B',
        }
    }

    NEW: {
        'step_version': 2, 
        'step_type': "set_column_formula",
        'params': {
            sheet_index: 0,
            column_id: _column id_,
            old_formula: '=A',
            new_formula: '=B',
        }
    }
    """
    params = step['params']
    params = replace_headers_with_id(params, 'column_header', 'column_id')

    return [{
        "step_version": 2, 
        "step_type": "set_column_formula", 
        "params": params
    }] + later_steps
    
def upgrade_set_column_formula_2_to_3(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Adds in the formula_label, which defaults to 0 for all upgrades. This is used for calculating offsets 
    when users are writing rolling window formulas.

    OLD: {
        'step_version': 2, 
        'step_type': "set_column_formula",
        'params': {
            sheet_index: 0,
            column_id: _column id_,
            old_formula: '=A',
            new_formula: '=B',
        }
    }

    NEW: {
        'step_version': 3, 
        'step_type': "set_column_formula",
        'params': {
            sheet_index: 0,
            column_id: _column id_,
            formula_label: 0,
            old_formula: '=A',
            new_formula: '=B',
        }
    }
    """
    params = step['params']
    params['formula_label'] = 0

    return [{
        "step_version": 3, 
        "step_type": "set_column_formula", 
        "params": params
    }] + later_steps

def upgrade_set_column_formula_3_to_4(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Adds in the index_labels_formula_is_applied_to, which specifies how the formula is applied. This defaults
    to {'type': 'entire_column'}.

    OLD: {
        'step_version': 3, 
        'step_type': "set_column_formula",
        'params': {
            sheet_index: 0,
            column_id: _column id_,
            formula_label: 0,
            old_formula: '=A',
            new_formula: '=B',
        }
    }

    NEW: {
        'step_version': 4, 
        'step_type': "set_column_formula",
        'params': {
            sheet_index: 0,
            column_id: _column id_,
            formula_label: 0,
            index_labels_formula_is_applied_to: {'type': 'entire_column'}
            old_formula: '=A',
            new_formula: '=B',
        }
    }
    """
    params = step['params']
    params['index_labels_formula_is_applied_to'] = {'type': 'entire_column'}

    return [{
        "step_version": 4, 
        "step_type": "set_column_formula", 
        "params": params
    }] + later_steps

def upgrade_set_column_formula_4_to_5(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Adds public interface version.

    OLD: {
        'step_version': 4, 
        'step_type': "set_column_formula",
        'params': {
            sheet_index: 0,
            column_id: _column id_,
            formula_label: 0,
            index_labels_formula_is_applied_to: {'type': 'entire_column'}
            old_formula: '=A',
            new_formula: '=B',
        }
    }
    NEW: {
        'step_version': 5, 
        'step_type': "set_column_formula",
        'params': {
            sheet_index: 0,
            column_id: _column id_,
            formula_label: 0,
            index_labels_formula_is_applied_to: {'type': 'entire_column'}
            old_formula: '=A',
            new_formula: '=B',
            public_interface_version: 1,
        }
    }
    """
    params = step['params']
    params['public_interface_version'] = 1

    return [{
        "step_version": 5, 
        "step_type": "set_column_formula", 
        "params": params
    }] + later_steps



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.


from copy import copy
from typing import TYPE_CHECKING, List, Optional, Any, Type
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.column_steps.delete_column_code_chunk import DeleteColumnsCodeChunk
from mitosheet.code_chunks.step_performers.filter_code_chunk import FilterCodeChunk
from mitosheet.code_chunks.step_performers.sort_code_chunk import SortCodeChunk
from mitosheet.pro.code_chunks.code_chunk_pro_utils import optimize_code_chunks

if TYPE_CHECKING:
    from mitosheet.step import Step
else:
    Step = Any
    

def get_code_chunks(all_steps: List[Step], optimize: bool=True) -> List[CodeChunk]:
    """
    A utility for taking all the steps in the steps manager, and returning a list
    of CodeChunks that correspond to these steps. 

    optimize is by default True, which results in these CodeChunks being optimized
    down to the smallest possible list of CodeChunks that implements the same ops.
    """
    from mitosheet.steps_manager import get_step_indexes_to_skip
    step_indexes_to_skip = get_step_indexes_to_skip(all_steps)

    all_code_chunks: List[CodeChunk] = []
    for step_index, step in enumerate(all_steps):
        # Skip the initalize step, or any step we should skip
        if step.step_type == 'initialize' or step_index in step_indexes_to_skip:
            continue

        all_code_chunks.extend(step.step_performer.transpile(
            step.prev_state, # type: ignore
            step.params,
            step.execution_data,
        ))

    if optimize:
        code_chunks_list = optimize_code_chunks(all_code_chunks)
    else:
        code_chunks_list = all_code_chunks

    return code_chunks_list


# NOTE: we cannot use get_right_combine_with_column_delete_code_chunk on sort/filter, 
# as sort potentially changes the indexes of the dataframe, which is a lasting change
# that occurs even after this column is deleted. Hence, we throw errors in this util 
# so that remember to not do that
CANNOT_RIGHT_COMBINE_WITH_DELETE_COLUMNS: List[Type[CodeChunk]] = [
    FilterCodeChunk,
    SortCodeChunk
]

def get_right_combine_with_column_delete_code_chunk(
        code_chunk: CodeChunk,
        delete_columns_code_chunk: DeleteColumnsCodeChunk,
        sheet_index_key: str, 
        column_id_key: str, 
    ) -> Optional[CodeChunk]:
    """
    A lot of the operations that work on single columns, when right combined with
    a delete column chunk, can be optimized out. 

    As such, we have a general utility we can apply to apply this optimization
    to a variety of steps, including: setting cell values, setting column formulas.

    NOTE: If you set a cell value, and there are formulas dependent on this column,
    then these formulas refresh. Thus, it might seem like optimizing out the setting
    cell values might lead to invalid code - but notably, you cannot delete columns
    that have columns that are dependant on them - so this is not a problem!
    """
    for invalid_code_chunk_type in CANNOT_RIGHT_COMBINE_WITH_DELETE_COLUMNS:
        if isinstance(code_chunk, invalid_code_chunk_type):
            raise Exception("Code chunk of this type is not valid for right combine with delete", CANNOT_RIGHT_COMBINE_WITH_DELETE_COLUMNS)

    # If the sheet indexes don't match, bail
    if code_chunk.__dict__[sheet_index_key] != delete_columns_code_chunk.sheet_index:
        return None

    column_id = code_chunk.__dict__[column_id_key]
    deleted_column_ids = delete_columns_code_chunk.column_ids

    if column_id in deleted_column_ids:
        new_deleted_column_ids = copy(deleted_column_ids)
        new_deleted_column_ids.remove(column_id)

        return DeleteColumnsCodeChunk(
            code_chunk.prev_state,
            delete_columns_code_chunk.sheet_index,
            delete_columns_code_chunk.column_ids
        )
    
    return None



import json
import pprint
import time
from typing import Any, Optional, List, Dict
import requests

def preprocess_log_for_upload(log_event: str, log_params: Dict[str, Any]) -> Optional[Dict[str, Any]]:
    """
    Convert logs into the correct format and remove any log events that are not part of the whitelisted schema.

	{
	    'timestamp': '2023-10-25T15:30:00Z',
	    'event': 'set_column_formula',
	    'params': {
		'sheet_index': 1,
		'column_id': 'column id',
		'new_formula': '=10 + 11'
	     },
	     'version_python': '3.9',
	     'version_pandas': '2.0',
	     'version_mitosheet': '0.1.522',
	}
    """
    
    whitelisted_log_events = [
        'edit_event',
        'error', 
        'mitosheet_rendered'
    ]

    whitelisted_log_params = [
        'version_python',
        'version_pandas',
        'version_mito',
        'error_traceback',
        'error_traceback_last_line',
    ]

    # Remove non-whitelisted events
    if log_event not in whitelisted_log_events:
        return None

    # Remove any log params that are not part of whitelisted params or start with "params", ie: params_sheet_index
    filtered_log_params = {k: v for k, v in log_params.items() if k in whitelisted_log_params or k.startswith('params_')}

    # Add the gmt timestamp formatted as 2023-10-25T15:30:00Z
    filtered_log_params['timestamp_gmt'] = time.strftime('%Y-%m-%dT%H:%M:%SZ', time.gmtime())

    # Create a useful top-level event field. Use the params_log_event if it exists (edit_events)
    # Otherwise, use the log_event.
    if 'params_log_event' in filtered_log_params:
        filtered_log_params['event'] = filtered_log_params['params_log_event']
        del filtered_log_params['params_log_event']
    else:
        filtered_log_params['event'] = log_event

    return filtered_log_params


class MitoLogUploader:
    """
    The MitoLogUploader is responsible for uploading logs to a 
    custom analytics url set in the MITO_CONFIG.

    It only uploads the most useful logs and log params in order 
    to make the logs maximally useful and minimally confusing. 
     
    It also uploads logs in batches every log_interval seconds so that 
    enterprises collecting logs from thousands of users do not 
    break their logging server. 
    """
    def __init__(
        self, 
        log_url: str,
        log_interval: Optional[int]
    ):
        self.log_url = log_url
        self.log_interval = log_interval if log_interval is not None else 10
        self.last_upload_time = time.time()
        self.unprocessed_logs: List[Dict[str, Any]] = []

    def log(self, log_event: str, log_params: Dict[str, Any]) -> None:
        """
        Converts log into the correct format, adds it to the queue of logs to be uploaded,
        and checks if it is time to upload the logs.
        """
        filtered_log_params = preprocess_log_for_upload(log_event, log_params)

        if filtered_log_params is not None:
            self.unprocessed_logs.append(filtered_log_params)

        current_time = time.time()
        if self.last_upload_time + self.log_interval < current_time and len(self.unprocessed_logs) > 0:
            self.upload_log(current_time)

    def upload_log(self, last_processed_log_time: float) -> None:
        """
        Uploads the unprocessed logs to the log_url and clears the unprocessed logs.
        """
        log_payload = json.dumps(self.unprocessed_logs)
        self.last_upload_time = last_processed_log_time

        try:
            requests.post(
                self.log_url,
                data=log_payload,
                headers={'Content-Type': 'application/json'}
            )

            self.unprocessed_logs = []

        except Exception as e:
            print("Log upload failed with error: ", e)




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from typing import Any, Dict, List, Optional, Set, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.merge_code_chunk import \
    MergeCodeChunk
from mitosheet.errors import make_incompatible_merge_key_error
from mitosheet.state import DATAFRAME_SOURCE_MERGED, State
from mitosheet.step_performers.pivot import get_optional_code_to_replay_on_dataframe_creation
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils.utils import get_param
from mitosheet.types import ColumnID, StepType
from mitosheet.utils import get_first_unused_dataframe_name

LOOKUP = 'lookup'
UNIQUE_IN_LEFT = 'unique in left'
UNIQUE_IN_RIGHT = 'unique in right'

class MergeStepPerformer(StepPerformer):
    """
    Allows you to merge two dataframes together.
    """

    @classmethod
    def step_version(cls) -> int:
        return 5

    @classmethod
    def step_type(cls) -> str:
        return 'merge'
    
    @classmethod
    def saturate(cls, prev_state: State, params: Dict[str, Any], previous_steps: List[StepType]) -> Dict[str, Any]:

        # Replay the edits on top of the merged dataframe
        optional_code, optional_code_chunk_names = get_optional_code_to_replay_on_dataframe_creation(previous_steps, params)
        params['optional_code'] = optional_code
        params['optional_code_chunk_names'] = optional_code_chunk_names

        return params

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        destination_sheet_index: Optional[int] = get_param(params, 'destination_sheet_index')
        
        final_destination_sheet_index = destination_sheet_index if destination_sheet_index is not None else len(prev_state.dfs)

        new_df_name = get_first_unused_dataframe_name(prev_state.df_names, "df_merge") if destination_sheet_index is None else prev_state.df_names[destination_sheet_index]

        execution_data = {
            # NOTE: we return the final destination name so frontend merges can match on this execution data - it's not
            # actually used by the backend (which is kinda confusing and should be renamed)
            'destination_sheet_index': final_destination_sheet_index,
            'new_df_name': new_df_name,
        }

        try:
            return cls.execute_through_transpile(
                prev_state, 
                params, 
                execution_data,
                new_dataframe_params={
                    'df_source': DATAFRAME_SOURCE_MERGED,
                    'new_df_names': [new_df_name],
                    'overwrite': {
                        'sheet_index_to_overwrite': destination_sheet_index,
                        'attempt_to_save_filter_metadata': True
                    } if destination_sheet_index is not None else destination_sheet_index
                },
                optional_code=params.get('optional_code')
            )

        except ValueError:

            sheet_index_one: int = get_param(params, 'sheet_index_one')
            sheet_index_two: int = get_param(params, 'sheet_index_two')
            merge_key_column_ids: List[List[ColumnID]] = get_param(params, 'merge_key_column_ids')
            merge_keys_one = prev_state.column_ids.get_column_headers_by_ids(sheet_index_one, list(map(lambda x: x[0], merge_key_column_ids)))
            merge_keys_two = prev_state.column_ids.get_column_headers_by_ids(sheet_index_two, list(map(lambda x: x[1], merge_key_column_ids)))

            # If we get a value error from merging two incompatible columns, we go through and check 
            # to see which of the columns this is, so our error can be maximally informative
            for merge_key_one, merge_key_two in zip(merge_keys_one, merge_keys_two):
                merge_key_one_dtype = str(prev_state.dfs[sheet_index_one][merge_key_one].dtype)
                merge_key_two_dtype = str(prev_state.dfs[sheet_index_two][merge_key_two].dtype)

                if merge_key_one_dtype != merge_key_two_dtype:
                    raise make_incompatible_merge_key_error(
                        merge_key_one=merge_key_one, 
                        merge_key_one_dtype=merge_key_one_dtype,
                        merge_key_two=merge_key_two, 
                        merge_key_two_dtype=merge_key_two_dtype,
                        error_modal=False
                    )

            raise make_incompatible_merge_key_error(error_modal=False)


    @classmethod
    def transpile(
        cls,
        prev_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            MergeCodeChunk(
                prev_state, 
                get_param(params, 'how'),
                get_param(params, 'destination_sheet_index'),
                get_param(params, 'sheet_index_one'),
                get_param(params, 'sheet_index_two'),
                get_param(params, 'merge_key_column_ids'),
                get_param(params, 'selected_column_ids_one'),
                get_param(params, 'selected_column_ids_two'),
                get_param(execution_data if execution_data is not None else {}, 'new_df_name'),
                get_param(execution_data if execution_data is not None else {}, 'optional_code_that_successfully_executed')
            )
        ]
    
    
    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        destination_sheet_index = get_param(params, 'destination_sheet_index')
        if destination_sheet_index is not None: # If editing an existing sheet, that is what is changed
            return {destination_sheet_index}
        return {-1}